/*
** Module   : BrowseInterface.java
** Abstract : Define all BROWSE-specific widget attributes and methods available via a handle
**            instance.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SVL 20141106 Created initial version.
** 002 SVL 20141209 Updated javadoc.
** 003 GES 20141124 Moved scrollbar-vertical to a separate interface since it is not browse-only.
** 004 SVL 20150503 Added addLikeColumn(String, integer).
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * Define all BROWSE-specific widget attributes and methods which can be accessed via a handle
 * instance.  Attributes and methods shared by multiple widget types will be defined elsewhere.
 */
public interface BrowseInterface
{
   /**
    * Implements the FETCH-SELECTED-ROW() widget method.
    * 
    * @param   n
    *          1-based index of the selected row to fetch;  note that this is
    *          the index into the list of selected rows within the browse,
    *          NOT the index of the row itself.
    *
    * @return  <code>true</code> if method succeeds.
    */
   @LegacyMethod(name = "FETCH-SELECTED-ROW")
   public logical fetchSelectedRow(int64 n);
   
   /**
    * Implements the FETCH-SELECTED-ROW() widget method.
    * 
    * @param   n
    *          1-based index of the selected row to fetch;  note that this is
    *          the index into the list of selected rows within the browse,
    *          NOT the index of the row itself.
    *
    * @return  <code>true</code> if method succeeds.
    */
   @LegacyMethod(name = "FETCH-SELECTED-ROW")
   public logical fetchSelectedRow(int n);
   
   /**
    * Get value of the FIT-LAST-COLUMN attribute.
    *
    * @return  value of the attribute.
    */
   @LegacyAttribute(name = "FIT-LAST-COLUMN")
   public logical isFitLastColumn();

   /**
    * Set value of the FIT-LAST-COLUMN attribute.
    *
    * @param    fitLastColumn
    *           New value of the attribute.
    */
   @LegacyAttribute(name = "FIT-LAST-COLUMN", setter = true)
   public void setFitLastColumn(boolean fitLastColumn);

   /**
    * Set value of the FIT-LAST-COLUMN attribute.
    *
    * @param    fitLastColumn
    *           New value of the attribute.
    */
   @LegacyAttribute(name = "FIT-LAST-COLUMN", setter = true)
   public void setFitLastColumn(logical fitLastColumn);

   /**
    * Set value of the EXPANDABLE attribute.
    *
    * @param    expandable
    *           New value of the attribute.
    */
   @LegacyAttribute(name = "EXPANDABLE", setter = true)
   public void setExpandable(boolean expandable);

   /**
    * Set value of the EXPANDABLE attribute.
    *
    * @param    expandable
    *           New value of the attribute.
    */
   @LegacyAttribute(name = "EXPANDABLE", setter = true)
   public void setExpandable(logical expandable);

   /**
    * Get value of the EXPANDABLE attribute.
    *
    * @return    value of the attribute.
    */
   @LegacyAttribute(name = "EXPANDABLE")
   public logical isExpandable();

   /**
    * Set value of the COLUMN-RESIZABLE attribute.
    *
    * @param    resizable
    *           <code>true</code> you can resize a browse column by pointing, clicking, and
    *           dragging.
    */
   @LegacyAttribute(name = "COLUMN-RESIZABLE", setter = true)
   public void setColumnResizable(boolean resizable);

   /**
    * Set value of the COLUMN-RESIZABLE attribute.
    *
    * @param    resizable
    *           <code>true</code> you can resize a browse column by pointing, clicking, and
    *           dragging.
    */
   @LegacyAttribute(name = "COLUMN-RESIZABLE", setter = true)
   public void setColumnResizable(logical resizable);

   /**
    * Get value of the COLUMN-RESIZABLE attribute.
    *
    * @return   <code>true</code> you can resize a browse column by pointing, clicking, and
    *           dragging.
    */
   @LegacyAttribute(name = "COLUMN-RESIZABLE")
   public logical isColumnResizable();

   /**
    * Set value of the COLUMN-MOVABLE attribute.
    *
    * @param    movable
    *           <code>true</code> you can move a browse column by pointing, clicking, and
    *           dragging.
    */
   @LegacyAttribute(name = "COLUMN-MOVABLE", setter = true)
   public void setColumnMovable(boolean movable);

   /**
    * Set value of the COLUMN-MOVABLE attribute.
    *
    * @param    movable
    *           <code>true</code> you can move a browse column by pointing, clicking, and
    *           dragging.
    */
   @LegacyAttribute(name = "COLUMN-MOVABLE", setter = true)
   public void setColumnMovable(logical movable);

   /**
    * Get value of the COLUMN-MOVABLE attribute.
    *
    * @return   <code>true</code> you can move a browse column by pointing, clicking, and
    *           dragging.
    */
   @LegacyAttribute(name = "COLUMN-MOVABLE")
   public logical isColumnMovable();

   /**
    * Get value of the CREATE-ON-ADD attribute.
    *
    * @return   value of the attribute.
    */
   @LegacyAttribute(name = "CREATE-ON-ADD")
   public logical isCreateOnAdd();

   /**
    * Set value of the CREATE-ON-ADD attribute.
    *
    * @param    createOnAdd
    *           New value of the attribute.
    */
   @LegacyAttribute(name = "CREATE-ON-ADD", setter = true)
   public void setCreateOnAdd(boolean createOnAdd);

   /**
    * Set value of the CREATE-ON-ADD attribute.
    *
    * @param    createOnAdd
    *           New value of the attribute.
    */
   @LegacyAttribute(name = "CREATE-ON-ADD", setter = true)
   public void setCreateOnAdd(logical createOnAdd);

   /**
    * Set value of the REFRESHABLE attribute.
    *
    * @param    refreshable
    *           <code>true</code> if the rows that appear in a browse are refreshed when an
    *           application opens or repositions a query.
    */
   @LegacyAttribute(name = "REFRESHABLE", setter = true)
   public void setRefreshable(boolean refreshable);

   /**
    * Set value of the REFRESHABLE attribute.
    *
    * @param    refreshable
    *           <code>true</code> if the rows that appear in a browse are refreshed when an
    *           application opens or repositions a query.
    */
   @LegacyAttribute(name = "REFRESHABLE", setter = true)
   public void setRefreshable(logical refreshable);

   /**
    * Get value of the REFRESHABLE attribute.
    *
    * @return   <code>true</code> if the rows that appear in a browse are refreshed when an
    *           application opens or repositions a query.
    */
   @LegacyAttribute(name = "REFRESHABLE")
   public logical isRefreshable();

   /**
    * Get value of the CURRENT-ROW-MODIFIED attribute.
    *
    * @return   <code>true</code> if any cells in the current row have been changed.
    */
   @LegacyAttribute(name = "CURRENT-ROW-MODIFIED")
   public logical isCurrentRowModified();

   /**
    * Get value of the NEW-ROW attribute.
    *
    * @return   <code>true</code> if the row in focus was added to the browse using the
    *           INSERT-ROW() method and has not been added to the database.
    */
   @LegacyAttribute(name = "NEW-ROW")
   public logical isNewRow();

   /**
    * Get value of the ROW-MARKERS attribute.
    *
    * @return   <code>true</code> if the browse has row markers.
    */
   @LegacyAttribute(name = "ROW-MARKERS")
   public logical hasRowMarkers();

   /**
    * Gets the LABEL-BGCOLOR writable attribute.
    *
    * @return   The current value of the LABEL-BGCOLOR attribute.
    */
   @LegacyAttribute(name = "LABEL-BGCOLOR")
   public integer getLabelBgColor();

   /**
    * Sets the LABEL-BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the LABEL-BGCOLOR attribute.
    */
   @LegacyAttribute(name = "LABEL-BGCOLOR", setter = true)
   public void setLabelBgColor(integer bgcolor);

   /**
    * Sets the LABEL-BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the LABEL-BGCOLOR attribute.
    */
   @LegacyAttribute(name = "LABEL-BGCOLOR", setter = true)
   public void setLabelBgColor(int bgcolor);

   /**
    * Set value of the ROW-HEIGHT-CHARS attribute.
    *
    * @param    rowHeightChars
    *           New value of the attribute.
    */
   @LegacyAttribute(name = "ROW-HEIGHT-CHARS", setter = true)
   public void setRowHeightChars(double rowHeightChars);

   /**
    * Set value of the ROW-HEIGHT-CHARS attribute.
    *
    * @param    rowHeightChars
    *           New value of the attribute.
    */
   @LegacyAttribute(name = "ROW-HEIGHT-CHARS", setter = true)
   public void setRowHeightChars(NumberType rowHeightChars);

   /**
    * Implements the ROW-HEIGHT-CHARS attribute getter.<p>
    *
    * @return row height, in characters, of the browse.
    */
   @LegacyAttribute(name = "ROW-HEIGHT-CHARS")
   public decimal getRowHeightChars();

   /**
    * Implements the SELECT-FOCUSED-ROW() widget method.<p>
    * Selects the row that currently has focus in a browse widget, even if it is not currently
    * displayed.
    *
    * @return   Always <code>true</code>.
    */
   @LegacyMethod(name = "SELECT-FOCUSED-ROW")
   public logical selectFocusedRow();

   /**
    * Implements the DESELECT-FOCUSED-ROW() widget method.<p>
    * Deselects the row with current focus.
    *
    * @return   <code>true</code> if the browse has at least one record.
    */
   @LegacyMethod(name = "DESELECT-FOCUSED-ROW")
   public logical deselectFocusedRow();

   /**
    * Implements the SELECT-ROW() widget method.<p>
    * Selects the specified row if it is currently in the browse viewport. In a single-select
    * browse, the previously selected row is deselected. No rows are deselected in a
    * multiple-select browse.
    *
    * @param   row
    *          1-based row index in the browse viewport.
    *
    * @return  <code>true</code> if the row with the given index exist in the browse viewport.
    */
   @LegacyMethod(name = "SELECT-ROW")
   public logical selectRow(integer row);

   /**
    * Implements the SELECT-ROW() widget method.<p>
    * Selects the specified row if it is currently in the browse viewport. In a single-select
    * browse, the previously selected row is deselected. No rows are deselected in a
    * multiple-select browse.
    *
    * @param   row
    *          1-based row index in the browse viewport.
    *
    * @return  <code>true</code> if the row with the given index exist in the browse viewport.
    */
   @LegacyMethod(name = "SELECT-ROW")
   public logical selectRow(int row);

   /**
    * Implements the SELECT-NEXT-ROW() widget method.<p>
    * Deselects all currently selected rows in a browse and selects the row after the deselected
    * row.<p>
    * This method also repositions the query to the new row and copies the record into the
    * database buffer. The browse automatically scrolls to the selected row if it is out of
    * view.<p>
    * This method is intended for use with a browse that supports the selection of a single row at
    * a time (MULTIPLE attribute is set to FALSE). If more than one row is selected when you
    * execute this method, all of the selected rows are deselected and the record after the last
    * selected row becomes the selected row.
    *
    * @return  TODO
    */
   @LegacyMethod(name = "SELECT-NEXT-ROW")
   public logical selectNextRow();

   /**
    * Implements the SELECT-PREV-ROW() widget method.<p>
    * Deselects all currently selected rows in a browse and selects the row before the deselected
    * row.<p>
    * This method also repositions the query to the new row and copies the record into the
    * database buffer.The browse automatically scrolls to the selected row if it is out of
    * view.<p>
    * This method is intended for use with a browse that supports the selection of a single row at
    * a time (MULTIPLE attribute is set to FALSE). If more than one row is selected when you
    * execute this method, all of the selected rows are deselected and the record before the last
    * selected row becomes the selected row.
    *
    * @return  TODO
    */
   @LegacyMethod(name = "SELECT-PREV-ROW")
   public logical selectPreviousRow();

   /**
    * Implements the DESELECT-SELECTED-ROW() widget method.<p>
    * Deselects the <i>n</i>th selected row in a browse.<p>
    * Progress maintains a numbered list of selected rows, starting at 1. When the
    * DESELECT-SELECTED-ROW(n) method is encountered, Progress searches this list to find
    * the <i>n</i>th selected row.
    *
    * @param   row
    *          1-based index of a selected row (rows are numbered in the order they were
    *          selected).
    *
    * @return  <code>true</code> if the row is successfully deselected.
    */
   @LegacyMethod(name = "DESELECT-SELECTED-ROW")
   public logical deselectSelectedRow(integer row);

   /**
    * Implements the DESELECT-SELECTED-ROW() widget method.<p>
    * Deselects the <i>n</i>th selected row in a browse.<p>
    * Progress maintains a numbered list of selected rows, starting at 1. When the
    * DESELECT-SELECTED-ROW(n) method is encountered, Progress searches this list to find
    * the <i>n</i>th selected row.
    *
    * @param   row
    *          1-based index of a selected row (rows are numbered in the order they were
    *          selected).
    *
    * @return  <code>true</code> if the row is successfully deselected.
    */
   @LegacyMethod(name = "DESELECT-SELECTED-ROW")
   public logical deselectSelectedRow(int row);

   /**
    * Implements the DELETE-SELECTED-ROW() widget method.<p>
    * Deletes the <i>n</i>th selected row from a browse and the results list.<p>
    * Progress maintains a numbered list of selected rows, starting at 1. When the
    * DELETE-SELECTED-ROW(n) method is encountered, Progress searches this list to find
    * the <i>n</i>th selected row.<p>
    * This method does not delete the record from the database and has no effect on the database
    * buffer.
    *
    * @param   row
    *          1-based index of a row to delete (rows are numbered in the order they were
    *          selected).
    *
    * @return  <code>true</code> if the row is successfully deleted.
    */
   @LegacyMethod(name = "DELETE-SELECTED-ROW")
   public logical deleteSelectedRow(integer row);

   /**
    * Implements the DELETE-SELECTED-ROW() widget method.<p>
    * Deletes the <i>n</i>th selected row from a browse and the results list.<p>
    * Progress maintains a numbered list of selected rows, starting at 1. When the
    * DELETE-SELECTED-ROW(n) method is encountered, Progress searches this list to find
    * the <i>n</i>th selected row.<p>
    * This method does not delete the record from the database and has no effect on the database
    * buffer.
    *
    * @param   row
    *          1-based index of a row to delete (rows are numbered in the order they were
    *          selected).
    *
    * @return  <code>true</code> if the row is successfully deleted.
    */
   @LegacyMethod(name = "DELETE-SELECTED-ROW")
   public logical deleteSelectedRow(int row);

   /**
    * Implements the DELETE-SELECTED-ROWS() widget method.<p>
    * Deletes all currently selected rows from a browse and the associated results list.<p>
    * This method does not delete the record from the database and has no effect on the database
    * buffer.
    *
    * @return  <code>true</code> if rows were successfully deleted.
    */
   @LegacyMethod(name = "DELETE-SELECTED-ROWS")
   public logical deleteSelectedRows();

   /**
    * Implements the DELETE-CURRENT-ROW() widget method.<p>
    * Deletes the most recently selected row from a browse and the results list.<p>
    * This method does not delete the record from the database and has no effect on the database
    * buffer.
    *
    * @return  <code>true</code> if the row is successfully deleted.
    */
   @LegacyMethod(name = "DELETE-CURRENT-ROW")
   public logical deleteCurrentRow();

   /**
    * Implements the IS-ROW-SELECTED() widget method.<p>
    * Returns <code>true</code> if a specified row in the browse viewport is currently selected.
    *
    * @param   row
    *          1-based row index in the browse viewport.
    *
    * @return  <code>true</code> if a specified row in the browse viewport is currently selected.
    */
   @LegacyMethod(name = "IS-ROW-SELECTED")
   public logical isRowSelected(integer row);

   /**
    * Implements the IS-ROW-SELECTED() widget method.<p>
    * Returns <code>true</code> if a specified row in the browse viewport is currently selected.
    *
    * @param   row
    *          1-based row index in the browse viewport.
    *
    * @return  <code>true</code> if a specified row in the browse viewport is currently selected.
    */
   @LegacyMethod(name = "IS-ROW-SELECTED")
   public logical isRowSelected(int row);

   /**
    * Implements the INSERT-ROW() widget method.<p>
    * Inserts a blank line in an updateable browse before or after the last selected row. The
    * blank line is a placeholder for a new record to be added through the browse. This method
    * cannot be used with a read-only browse.
    *
    * @param   mode
    *          <code>"BEFORE"</code> if a row should be added before the current browse row.
    *          <code>"AFTER"</code> if a row should be added after the current browse row.
    *
    * @return  <code>true</code> if a row was successfully added.
    */
   @LegacyMethod(name = "INSERT-ROW")
   public logical insertRow(character mode);

   /**
    * Implements the INSERT-ROW() widget method.<p>
    * Inserts a blank line in an updateable browse before or after the last selected row. The
    * blank line is a placeholder for a new record to be added through the browse. This method
    * cannot be used with a read-only browse.
    *
    * @param   mode
    *          <code>"BEFORE"</code> if a row should be added before the current browse row.
    *          <code>"AFTER"</code> if a row should be added after the current browse row.
    *
    * @return  <code>true</code> if a row was successfully added.
    */
   @LegacyMethod(name = "INSERT-ROW")
   public logical insertRow(String mode);

   /**
    * Implements the INSERT-ROW() widget method.<p>
    * Inserts a blank line in an updateable browse before the last selected row. The
    * blank line is a placeholder for a new record to be added through the browse. This method
    * cannot be used with a read-only browse.
    *
    * @return  <code>true</code> if a row was successfully added.
    */
   @LegacyMethod(name = "INSERT-ROW")
   public logical insertRow();

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field and returns its widget handle.
    *
    * @param  fieldName
    *         The name of a field in one of the buffers associated with the browse’s query. If the
    *         query is a join, the name must be qualified with the database name.
    * @param  position
    *         1-based position of the browse column.
    *
    * @return handle to the created column.
    */
   @LegacyMethod(name = "ADD-LIKE-COLUMN")
   public handle addLikeColumn(String fieldName, integer position);

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field and returns its widget handle.
    *
    * @param  fieldName
    *         The name of a field in one of the buffers associated with the browse’s query. If the
    *         query is a join, the name must be qualified with the database name.
    * @param  position
    *         1-based position of the browse column.
    *
    * @return handle to the created column.
    */
   @LegacyMethod(name = "ADD-LIKE-COLUMN")
   public handle addLikeColumn(character fieldName, integer position);

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field and returns its widget handle.
    *
    * @param  fieldName
    *         The name of a field in one of the buffers associated with the browse’s query. If the
    *         query is a join, the name must be qualified with the database name.
    * @param  position
    *         1-based position of the browse column.
    *
    * @return handle to the created column.
    */
   @LegacyMethod(name = "ADD-LIKE-COLUMN")
   public handle addLikeColumn(String fieldName, int position);

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field and returns its widget handle.
    *
    * @param  fieldHandle
    *         Handle to a field in one of the buffers associated with the browse’s query.
    * @param  position
    *         1-based position of the browse column.
    *
    * @return handle to the created column.
    */
   @LegacyMethod(name = "ADD-LIKE-COLUMN")
   public handle addLikeColumn(handle fieldHandle, integer position);

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field and returns its widget handle.
    *
    * @param  fieldHandle
    *         Handle to a field in one of the buffers associated with the browse’s query.
    * @param  position
    *         1-based position of the browse column.
    *
    * @return handle to the created column.
    */
   @LegacyMethod(name = "ADD-LIKE-COLUMN")
   public handle addLikeColumn(handle fieldHandle, int position);

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field, adds it to the end of the columns and
    * returns its widget handle.
    *
    * @param  fieldName
    *         The name of a field in one of the buffers associated with the browse’s query. If the
    *         query is a join, the name must be qualified with the database name.
    *
    * @return handle to the created column.
    */
   @LegacyMethod(name = "ADD-LIKE-COLUMN")
   public handle addLikeColumn(character fieldName);

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field, adds it to the end of the columns and
    * returns its widget handle.
    *
    * @param  fieldName
    *         The name of a field in one of the buffers associated with the browse’s query. If the
    *         query is a join, the name must be qualified with the database name.
    *
    * @return handle to the created column.
    */
   @LegacyMethod(name = "ADD-LIKE-COLUMN")
   public handle addLikeColumn(String fieldName);

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field, adds it to the end of the columns and
    * returns its widget handle.
    *
    * @param  fieldHandle
    *         Handle to a field in one of the buffers associated with the browse’s query.
    *
    * @return handle to the created column.
    */
   @LegacyMethod(name = "ADD-LIKE-COLUMN")
   public handle addLikeColumn(handle fieldHandle);

   /**
    * Implements the MOVE-COLUMN() widget method.<p>
    * Repositions a column in a browse widget.
    *
    * @param  src
    *         1-based position of the column to be moved.
    * @param  dest
    *         1-based position to which the column is moved.
    *
    * @return <code>true</code> if the column was successfully moved.
    */
   @LegacyMethod(name = "MOVE-COLUMN")
   public logical moveColumn(integer src, integer dest);

   /**
    * Implements the MOVE-COLUMN() widget method.<p>
    * Repositions a column in a browse widget.
    *
    * @param  src
    *         1-based position of the column to be moved.
    * @param  dest
    *         1-based position to which the column is moved.
    *
    * @return <code>true</code> if the column was successfully moved.
    */
   @LegacyMethod(name = "MOVE-COLUMN")
   public logical moveColumn(int src, int dest);

   /**
    * Implements the GET-REPOSITIONED-ROW() widget method.<p>
    * Returns the row index of the browse viewport where the REPOSITION TO ROWID (or
    * RECID) statement displays a repositioned record.
    *
    * @return the row index of the browse viewport where the REPOSITION TO ROWID (or
    *         RECID) statement displays a repositioned record
    */
   @LegacyMethod(name = "GET-REPOSITIONED-ROW")
   public integer getRepositionedRow();

   /**
    * Implements the FIRST-COLUMN attribute getter.<p>
    *
    * @return a handle to the first column in a browse widget, regardless of the value of its
    *         READ-ONLY attribute or its VISIBLE attribute.
    */
   @LegacyAttribute(name = "FIRST-COLUMN")
   public handle getFirstColumn();

   /**
    * Implements the FOCUSED-ROW attribute getter.<p>
    *
    * @return the 1-based index of the focused row in the viewport.
    */
   @LegacyAttribute(name = "FOCUSED-ROW")
   public integer getFocusedRow();

   /**
    * Implements the NUM-LOCKED-COLUMNS attribute getter.<p>
    *
    * @return the number of visible leading columns locked in a browse widget.
    */
   @LegacyAttribute(name = "NUM-LOCKED-COLUMNS")
   public integer getNumLockedColumns();

   /**
    * Implements the NUM-LOCKED-COLUMNS attribute setter.<p>
    *
    * @param num
    *        The number of visible leading columns locked in a browse widget.
    */
   @LegacyAttribute(name = "NUM-LOCKED-COLUMNS", setter = true)
   public void setNumLockedColumns(double num);

   /**
    * Implements the NUM-LOCKED-COLUMNS attribute setter.<p>
    *
    * @param num
    *        The number of visible leading columns locked in a browse widget.
    */
   @LegacyAttribute(name = "NUM-LOCKED-COLUMNS", setter = true)
   public void setNumLockedColumns(NumberType num);

   /**
    * Implements the MIN-HEIGHT-CHARS attribute getter.<p>
    *
    * @return  minimum height of the browse, in character units.
    */
   @LegacyAttribute(name = "MIN-HEIGHT-CHARS")
   public decimal getMinHeightChars();

   /**
    * Implements the ALLOW-COLUMN-SEARCHING attribute getter.<p>
    *
    * @return <code>true</code> if column searching is allowed for the browse.
    */
   @LegacyAttribute(name = "ALLOW-COLUMN-SEARCHING")
   public logical isAllowColumnSearching();

   /**
    * Implements the ALLOW-COLUMN-SEARCHING attribute setter.<p>
    *
    * @param allow
    *        <code>true</code> if column searching is allowed for the browse.
    */
   @LegacyAttribute(name = "ALLOW-COLUMN-SEARCHING", setter = true)
   public void setAllowColumnSearching(boolean allow);

   /**
    * Implements the ALLOW-COLUMN-SEARCHING attribute setter.<p>
    *
    * @param allow
    *        <code>true</code> if column searching is allowed for the browse.
    */
   @LegacyAttribute(name = "ALLOW-COLUMN-SEARCHING", setter = true)
   public void setAllowColumnSearching(logical allow);

   /**
    * Set value of the MAX-DATA-GUESS attribute.
    *
    * @param    maxDataGuess
    *           The estimated number of records in a browse query.
    */
   @LegacyAttribute(name = "MAX-DATA-GUESS", setter = true)
   public void setMaxDataGuess(double maxDataGuess);

   /**
    * Set value of the MAX-DATA-GUESS attribute.
    *
    * @param    maxDataGuess
    *           The estimated number of records in a browse query.
    */
   @LegacyAttribute(name = "MAX-DATA-GUESS", setter = true)
   public void setMaxDataGuess(NumberType maxDataGuess);

   /**
    * Get value of the MAX-DATA-GUESS attribute.
    *
    * @return   the estimated number of records in a browse query.
    */
   @LegacyAttribute(name = "MAX-DATA-GUESS")
   public integer getMaxDataGuess();

   /**
    * Provides the SET-REPOSITIONED-ROW() method which assigns the row number and type of
    * repositioning for the browse.
    * 
    * @param    row
    *           1-based row number where the new record is displayed
    * @param    cond
    *           If this parameter is "conditional", then method tries to
    *           reposition to the browse viewport existing row first
    * 
    * @return   The result of operation. It evaluates to <code>true</code> if
    *           operation was successful.
    */
   @LegacyMethod(name = "SET-REPOSITIONED-ROW")
   public logical setRepositionedRow(int64 row, String cond);
   
   /**
    * Provides the SET-REPOSITIONED-ROW() method which assigns the row number and type of
    * repositioning for the browse.
    * 
    * @param    row
    *           1-based row number where the new record is displayed
    * @param    cond
    *           If this parameter is "conditional", then method tries to
    *           reposition to the browse viewport existing row first
    * 
    * @return   The result of operation. It evaluates to <code>true</code> if
    *           operation was successful.
    */
   @LegacyMethod(name = "SET-REPOSITIONED-ROW")
   public logical setRepositionedRow(int row, String cond);
   
   /**
    * Gets the SCROLLABLE writeable attribute.
    *
    * @return   The current value of the SCROLLABLE attribute.
    */
   @LegacyAttribute(name = "SCROLLABLE")
   public logical isScrollable();

   /**
    * Sets the SCROLLABLE writeable attribute.
    *
    * @param    scroll
    *           The new value for the SCROLLABLE attribute.
    */
   @LegacyAttribute(name = "SCROLLABLE", setter = true)
   public void setScrollable(logical scroll);

   /**
    * Sets the SCROLLABLE writeable attribute.
    *
    * @param    scroll
    *           The new value for the SCROLLABLE attribute.
    */
   @LegacyAttribute(name = "SCROLLABLE", setter = true)
   public void setScrollable(boolean scroll);

   /**
    * Set SEPARATORS attribute.
    *
    * @param    separators
    *           New value for the attribute.
    */
   @LegacyAttribute(name = "SEPARATORS", setter = true)
   public void setSeparators(boolean separators);

   /**
    * Set SEPARATORS attribute.
    *
    * @param    separators
    *           New value for the attribute.
    */
   @LegacyAttribute(name = "SEPARATORS", setter = true)
   public void setSeparators(logical separators);

   /**
    * Implements the SEPARATORS attribute getter.<p>
    *
    * @return <code>true</code> if row and column separators of the browse widget are displayed.
    */
   @LegacyAttribute(name = "SEPARATORS")
   public logical isSeparators();
}
