/*
** Module   : BrowseImpl.java
** Abstract :
**
** Copyright (c) 2012-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SIY 20111124 Extracted CHUI-specific part from Browse.
** 002 CA  20121022 Added support for optional query and column-list
**                  clauses.
** 003 GES 20140627 Changed color attribute method names to match naming standards.
** 004 CA  20140805 Do not allow ChUI impl classes access to APIs outside of ChuiOutputManager.
** 005 HC  20140911 GUI coordinate support - data types refactoring.
** 006 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 007 HC  20141021 Implemented window move/position support.
** 008 SVL 20141209 Added support for NO-UNDERLINE and for locked columns.
** 009 SVL 20150212 Browse reworked.
** 010 HC  20150323 Removed business logic from config classes.
** 011 SVL 20150420 Removed invalid REFRESHABLE state check in draw(), added notifyColumnsUpdated.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * CHUI-specific portion of implementation of Progress BROWSE widget.
 */
@SuppressWarnings("unchecked")
public class BrowseImpl
extends Browse<ChuiOutputManager>
{
   /** The renderer for columns labels area. */
   private LabelColumnRenderer rndrLabels = null;
   
   /** The renderer for a data area with values. */
   private DataColumnRenderer rndrData = new DataColumnRenderer();
   
   /** The renderer for data lines which are outside data model. */
   private SpaceColumnRenderer rndrSpace = new SpaceColumnRenderer();
   
   /**
    * Constructs the CHUI-specific version of Browse widget.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public BrowseImpl(WidgetId id, BrowseConfig cfg)
   {
      super(id, cfg);
   }
   
   /**
    * Sets position of terminal's cursor. Redirects this positioning to the
    * in-place editor if it is present. Otherwise does nothing because BROWSE
    * itself has no purpose for the hardware cursor.
    */
   @Override
   public void drawCaret()
   {
      FillIn<ChuiOutputManager,?> editor = getActiveEditor();
      
      if (editor != null)
         editor.drawCaret();
   }

   /**
    * Draws the widget: border, header, selection marks and line, internal
    * FillIn etc.
    */
   @Override
   public void draw()
   {
      // if the model was not initialized yet (i.e. no query was specified),
      // initialize it now.
      if (browseModel == null)
      {
         setQuery(this.config.id.asInt());
      }

      Color clr = UiUtils.getDColor(this);
      colorCommon = clr;
      colorHSep = clr.withUnderline();
      colorCurrent = clr.withReverse();

      // if frame is not alive, do not generate events and clear browse
      // content
      Widget<ChuiOutputManager> frame = (Widget<ChuiOutputManager>) UiUtils.locateFrame(this);
      
      if (frame != null && !frame.isVisible())
         return;
      
      if (frame instanceof Frame && !((Frame<ChuiOutputManager>) frame).isAlive())
      {
         resetRowDisplayState();
         this.dcData = null;
         setCurrentRow(0);
         repaint();
      }
      else
      {
         // else, trigger events, before drawing
         triggerRowDisplayEvents();
      }

      // save the row from where drawing was started last time
      oldDisplayRow = getScrollRow();

      Point origin = screenLocation();
      
      // in chui char coordinates are always whole numbers
      int x = (int) origin.x;
      int y = (int) origin.y;

      ChuiOutputManager screen = screen();
      
      if (config.box)
      {
         screen.boxToScreen(origin, dimension(), clr);
         
         // in chui char coordinates are always whole numbers
         int width = (int) width();

         Color tcl = config.titleDColor;

         if (config.title != null && config.title.length() > 0)
         {
            String titleText = (config.title.length() > width) ?
                                     config.title.substring(0, width) : 
                                     config.title;
                                     
            int titleOffset = (width - titleText.length()) / 2;
            screen.at(x + titleOffset, y).append(titleText, tcl);
         }
                           
         ++x;
         ++y;
      }

      if (drawLabels)
      {
         if (rcLabels == null)
         {
            if (rndrLabels == null)
               rndrLabels = new LabelColumnRenderer();
            else
               rndrLabels.clear();

            renderAbstractLine(rndrLabels, -1);
            rndrLabels.printResult();
         }

         // ilLabelsHeight is rounded, the comparison is safe
         for (int i = 0; i < ilLabelsHeight; ++i)
         {
            screen.at(x, y).append(rcLabels[i], colorCommon);
            ++y;
         }

         if (!config.noUnderline)
         {
            int lockedColumnsSeparatorPos = 0;
            if (config.numLockedColumns > 0)
            {
               int availWidth = getAvailableWidth();
               if (availWidth > 0)
                  lockedColumnsSeparatorPos = (int) separatorLen - availWidth - 1;
            }

            if (lockedColumnsSeparatorPos > 0)
            {
               // If there are locked columns, there a space in underline appearing under
               // column labels, which separates locked columns from non-locked.
               screen.at(x, y).
                     append("  ", colorCommon).
                     hLine(lockedColumnsSeparatorPos, colorCommon).
                     append(" ", colorCommon).
                     hLine(separatorLen - lockedColumnsSeparatorPos - 1, colorCommon);
            }
            else
            {
               screen.at(x, y).append("  ", colorCommon).hLine(separatorLen, colorCommon);
            }

            // ilRightSpaceWidth is rounded, the comparison is safe
            if (ilRightSpaceWidth > 0)
            {
               // this cycle usually works only 2 times so it is more effective
               // than to generate special string
               for (int i = 0; i < ilRightSpaceWidth; ++i)
                  screen.append(' ', colorCommon);
            }

            ++y;
         }
      }

      hadDraw = true;
      
      int rowsLeft = getCacheSize();

      FillIn<ChuiOutputManager,?> editor = getActiveEditor();
      boolean readOnly = (editor == null || editor.config().readOnly);
      boolean focused = (hasFocus() && isEnabled()) || pseudoFocus;
      
      int rowsOnPageData = Math.min(ilPageRowsCount, rowsLeft);
      config.numIterations = rowsOnPageData;

      int index = 0;
      
      for (; rowsOnPageData > 0; --rowsOnPageData)
      {
         rndrData.newRow(readOnly, focused, getCachedRowNumber(index));
         screen.at(x, y);
         renderAbstractLine(rndrData, index);
         ++y;
         ++index;
      }
      
      for (int i = ilPageRowsCount - rowsLeft; i > 0; --i)
      {
         screen.at(x, y);
         int row = getCachedRowNumber(index);
         if (row == -1 && index == 0)
         {
            // highlight the first row of an empty browse
            row = 0;
         }
         rndrSpace.newRow(readOnly, focused, row);
         renderAbstractLine(rndrSpace, index);
         ++y;
         ++index;
      }
      
      if (!focused || editor == null)
      {
         screen.at(0, 0);
      }
      else
      {
         editor.draw();
      }
   }

   /**
    * Notify the browse that some columns have been added to it.
    */
   @Override
   public void notifyColumnsUpdated()
   {
      int newScrollRow = getCachedRowNumber(1);
      if (newScrollRow == -1)
      {
         // actually in this case in 4GL ChUI at this point browse is unusable
         newScrollRow = getCachedRowNumber(0);
      }

      browseModel.resetColumnsCache();
      rcLabels = null;
      rndrLabels = null;
      updateLabelsHeigth();

      WidgetRegistry widgetRegistry = ThinClient.getInstance().registry();
      BrowseColumnConfig[] cols = browseModel.getColumns();
      for (BrowseColumnConfig col : cols)
      {
         Widget component = widgetRegistry.getComponent(col.id);
         if (component == null)
         {
            widgetRegistry.reconstructWidget(col);
         }
      }

      super.notifyColumnsUpdated(false);

      // This is ChUI-only quirk
      if (newScrollRow != -1)
      {
         boolean editing = getActiveEditor() != null;
         if (editing)
         {
            undoValue();
            stopEdit();
         }

         scrollToRow(newScrollRow, true, false, true);

         for (int i = 0; i < cols.length; ++i)
         {
            if (!cols[i].readOnly)
            {
               currColumn = i;
               break;
            }
         }

         if (editing)
            startEdit();
      }
      else
      {
         refresh(false);
      }
   }

   /**
    * Renders a line from horizontally scrollable area.
    * 
    * @param renderer
    *        Implementation of various rendering's parts.
    * @param rowIdx
    *        The screen index of the row currently being displayed
    */
   private void renderAbstractLine(ColumnRenderer renderer, int rowIdx)
   {
      renderer.addMarks();
      BrowseColumnConfig[] cols = browseModel.getColumns();
      // columns may have not been specified, so exit early if so
      if (cols == null || cols.length == 0)
         return;

      // reset columns
      for (int i = 0; i < cols.length; i++)
         cols[i].column = -1;

      int availWidth = getAvailableWidth();

      if (scrollColumn == 0 || availWidth <= 0 || scrollColumn < config.numLockedColumns)
      {
         // no scrolling, the first column is fully visible
         // in chui char coordinates are always whole numbers
         int width = (int) ilPageWidth;
         
         BrowseColumnConfig col = cols[0];
         
         // in chui char coordinates are always whole numbers
         int colWidth = (int) Math.min(col.widthChars, width);

         if (rowIdx >= 0        &&
             colors != null     &&
             rowIdx < colors.length)
         {
            Color color = colors[rowIdx][0];
            col.columnDColor = color == null ? Color.DEFAULT : color;
         }
         renderer.addData(col, 0, colWidth, 0);
         width -= colWidth;
         int i;
         
         for (i = 1; i < cols.length; ++i)
         {
            if (width > 0)
            {
               renderer.addSeparator();
               --width;
            }

            if (width <= 0)
            {
               break;
            }
            
            col = cols[i];
            
            // in chui char coordinates are always whole numbers
            colWidth = (int) col.widthChars;

            if (width < colWidth && i == config.numLockedColumns)
            {
               // Special case: no scrolling, some locked columns and single non-locked
               // column which doesn't entirely fit into available width.
               colWidth = width;
            }

            int scroll = (colWidth > width) ? width - colWidth : 0;

            if (rowIdx >= 0        &&
                colors != null     &&
                rowIdx < colors.length)
            {
               Color color = colors[rowIdx][i];
               col.columnDColor = color == null ? Color.DEFAULT : color;
            }
            renderer.addData(col, i, colWidth, scroll);
            width -= colWidth;
         }
         
         config.numVisibleColumns = i;
      }
      else
      {
         int nCols = 1;

         // Draw locked columns
         if (config.numLockedColumns > 0)
         {
            for (int i = 0; i < config.numLockedColumns; i++)
            {
               BrowseColumnConfig col = cols[i];
               renderer.addData(col, i, (int) col.widthChars, 0);
               renderer.addSeparator();
               nCols++;
            }
         }

         // scrolling active, scroll column is on the right of view area
         BrowseColumnConfig scol = cols[scrollColumn];

         // in chui char coordinates are always whole numbers
         int scolWidth = (int) scol.widthChars;
         
         if (availWidth <= scolWidth)
         {
            // column is at least as wide as visible area
            scolWidth = availWidth;
         }
         else
         { 
            // view area contains also some columns before scroll column
            int fcolWidth = (int) (availWidth - scolWidth - 1);
            int fcolIndex = scrollColumn - 1;
            
            while (true)
            {
               // in chui char coordinates are always whole numbers
               int w = (int) (fcolWidth - cols[fcolIndex].widthChars - 1);
               if (w < 0)
               {
                  break;
               }

               fcolWidth = w;
               --fcolIndex;
            }

            BrowseColumnConfig col;
            
            if (fcolWidth > 0) // it may equal 0 also
            {
               col = cols[fcolIndex];
               int fullWidth = (int) col.widthChars;

               renderer.addData(col, fcolIndex, (int) fullWidth, (int) (fullWidth - fcolWidth));
               nCols++;
            }
            renderer.addSeparator();
            
            for (int i = fcolIndex + 1; i < scrollColumn; ++i)
            {
               col = cols[i];
               
               // in chui char coordinates are always whole numbers
               renderer.addData(col, i, (int) col.widthChars, 0);
               renderer.addSeparator();
               nCols++;
            }
         }

         renderer.addData(scol, scrollColumn, scolWidth, 0);
         config.numVisibleColumns = nCols;
      }
      
      // in chui char coordinates are always whole numbers
      if ((int) ilRightSpaceWidth > 0)
      {
         renderer.addRightSpaces((int) ilRightSpaceWidth);
      }
   }

   /**
    * Outputs several spaces to the terminal.
    *
    * @param   width
    *          Number of spaces.
    * @param   color
    *          Drawing color.
    */
   private void printSpaces(int width, Color color)
   {
      for (int i = width; i > 0; --i)
      {
         screen().append(' ', color);
      }
   }
   
   /**
    * Interface for renderers of horizontally scrollable rows. The place for
    * saving the result is up to implementation. The result may consist from
    * several character rows.
    */
   private interface ColumnRenderer
   {
      /**
       * Renders data from an one column.
       * 
       * @param   col
       *          Column to render.
       * @param   colIndex
       *          Index of the column.
       * @param   colWidth
       *          The full width for formatting.
       * @param   scroll
       *          Scrolling correction. Means the number of characters which
       *          should be erased from the left (if positive) or from the
       *          right (if negative). This scroll differs from the scroll in
       *          <code>DisplayFormat.formatScreenValue()</code> because it
       *          means the scroll of the whole browse column not the scroll
       *          of screen value inside the column.
       */
      public void addData(BrowseColumnConfig col, int colIndex, int colWidth, 
                          int scroll);

      /**
       * Renders an editable line and a selected line marks.
       */
      public void addMarks();

      /**
       * Renders a separator character.
       */
      public void addSeparator();

      /**
       * Renders several spaces to make right space column.
       * 
       * @param   width
       *          The number of spaces.
       */
      public void addRightSpaces(int width);
   }

   /**
    * Renders the labels area. Places the internal result into the internal
    * buffers. It is expected that <code>clear()</code> method is called
    * before rendering and <code>printResult()</code> is called to finish
    * it.
    */
   private class LabelColumnRenderer
   implements ColumnRenderer
   {
      /** Temporary buffers for render result. */
      private StringBuffer[] bufs;

      /**
       * Constructor. Initializes from the current labels. Object should be
       * recreated if labels are changed.
       */
      public LabelColumnRenderer()
      {
         bufs = new StringBuffer[ilLabelsHeight];
         for (int i = 0; i < bufs.length; ++i)
         {
            bufs[i] = new StringBuffer(ilLabelsHeight);
         }
      }

      /**
       * Clears an old content.
       */
      public void clear()
      {
         for (int i = 0; i < bufs.length; ++i)
         {
            bufs[i].setLength(0);
         }
      }

      /**
       * Copies buffers content into <code>rcLabels</code> member.
       */
      public void printResult()
      {
         rcLabels = new String[bufs.length];
         for (int i = 0; i < bufs.length; ++i)
         {
            rcLabels[i] = bufs[i].toString();
         }
      }

      /**
       * @param   colIndex
       *          Is not used.
       * @see  ColumnRenderer#addData(com.goldencode.p2j.ui.BrowseColumnConfig, int, int, int)
       */
      @Override
      public void addData(BrowseColumnConfig col, int colIndex, int colWidth, int scroll)
      {
         String[] label = getLabelVar(col);
         
         for (int i = 0; i < label.length; ++i)
         {
            String str = renderCell(label[i], col, colWidth, scroll);
            bufs[i].append(str);
         }

         for (int i = label.length; i < bufs.length; ++i)
         {
            StringBuffer buf = bufs[i];
            for (int j = colWidth - Math.abs(scroll); j > 0; --j)
            {
               buf.append(' ');
            }
         }
      }

      /**
       * @see ColumnRenderer#addMarks()
       */
      @Override
      public void addMarks()
      {
         addRightSpaces(2);
      }

      /**
       * @see ColumnRenderer#addSeparator()
       */
      @Override
      public void addSeparator()
      {
         addRightSpaces(1);
      }

      /**
       * @see ColumnRenderer#addRightSpaces(int)
       */
      @Override
      public void addRightSpaces(int width)
      {
         for (int i = 0; i < bufs.length; ++i)
         {
            for (int j = 0; j < width; ++j)
            {
               bufs[i].append(' ');
            }
         }
      }
   }

   /**
    * Renders a data line which is outside data model. Result is immediately
    * placed to the terminal. No row init necessary.
    */
   private class SpaceColumnRenderer
   implements ColumnRenderer
   {
      /** Indicates read-only mode. */
      protected boolean dcrReadOnly;
      /** Is widget focused. */
      protected boolean dcrFocused;
      /** The index of currently rendered row. */
      protected int dcrRowIndex;

      /** Dummy constructor. */
      public SpaceColumnRenderer(){}

      /**
       * @param   col
       *          Is not used.
       * @param   colIndex
       *          Is not used.
       *          
       * @see  ColumnRenderer#addData(com.goldencode.p2j.ui.BrowseColumnConfig,
       *       int, int, int)
       */
      @Override
      public void addData(BrowseColumnConfig col,
                          int colIndex,
                          int colWidth,
                          int scroll)
      {
         printSpaces(colWidth - Math.abs(scroll), getColor(colorCommon));
      }

      /**
       * Calculate row color.
       *
       * @param   color
       *          Base color.
       * @return  Calculated color.
       */
      public Color getColor(Color color)
      {
         return (dcrReadOnly && dcrRowIndex == Math.max(getCurrentRow(), 0) && dcrFocused)
                    ? Color.DEFAULT.withReverse()
                    : (config.separators ? color.withUnderline() : color);
      }

      /**
       * @see  ColumnRenderer#addMarks()
       */
      @Override
      public void addMarks()
      {
         char editMark;
         Color editMarkAttr = colorCommon;
         if (editPossible && config.rowMarkers)
         {
            editMark = '*';
            
            if (dcrReadOnly && 
                dcrRowIndex == Math.max(getCurrentRow(), 0) &&
                isEnabled())
            {
               editMarkAttr = colorCurrent;
            }
         }
         else
         {
            editMark = ' ';
         }

         screen().append(editMark, editMarkAttr);
         screen().append(isRowSelected(dcrRowIndex) ? '>' : ' ', colorCommon);
      }

      /**
       * @see  ColumnRenderer#addSeparator()
       */
      @Override
      public void addSeparator()
      {
         Color clr = config.separators && separatorQuirk ? colorHSep : colorCommon;
         char chr = config.separators ? ChuiOutputManager.UACS_VLINE : ' ';

         screen().append(chr, clr);
      }

      /**
       * @see  ColumnRenderer#addRightSpaces(int)
       */
      @Override
      public void addRightSpaces(int width)
      {
         printSpaces(width, colorCommon);
      }

      /**
       * Begins rendering of a new row.
       * 
       * @param   readOnly
       *          The widget in read-only mode.
       * @param   focused
       *          The widget is focused.
       * @param   rowIndex
       *          Index of the row in the browse model.
       */
      public void newRow(boolean readOnly, boolean focused, int rowIndex)
      {
         dcrFocused = focused;
         dcrReadOnly = readOnly;
         dcrRowIndex = rowIndex;
      }
   }

   /**
    * Renders a data line immediately to the terminal. <code>newRow()</code>
    * method should be called before rendering of each row.
    */
   private class DataColumnRenderer
   extends SpaceColumnRenderer
   {
      /** Data to render. */
      protected BaseDataType[] dcrRow;
      
      /**
       * Reference to deferred updates of current row. Always equals
       * <code>currRowUpdates</code> or <code>null</code>.
       */
      protected BaseDataType[] dcrRowUpdates;
      
      /** Dummy constructor. */
      public DataColumnRenderer(){}

      /**
       * @see  SpaceColumnRenderer#newRow(boolean, boolean, int)
       */
      @Override
      public void newRow(boolean readOnly, boolean focused, int rowIndex)
      {
         super.newRow(readOnly, focused, rowIndex);
         
         dcrRow = dcGetRow(rowIndex);
         dcrRowUpdates = (readOnly || dcrRowIndex != getCurrentRow())
                         ? null
                         : currRowUpdates;
      }

      /**
       * @see SpaceColumnRenderer#addData(com.goldencode.p2j.ui.BrowseColumnConfig, int, int, int)
       */
      @Override
      public void addData(BrowseColumnConfig col,
                          int colIndex,
                          int colWidth,
                          int scroll)
      {
         col.column = Coordinate.scale(screen().cursor().x - location().x);

         if (!dcrReadOnly && 
              dcrFocused  && 
              dcrRowIndex == Math.max(getCurrentRow(), 0) &&
              colIndex == currColumn)
         {
            // just cursor moving
            printSpaces(colWidth - Math.abs(scroll), colorCommon);
         }
         else
         {
            BaseDataType var =
               (dcrRowUpdates != null && dcrRowUpdates[colIndex] != null) ?
                  dcrRowUpdates[colIndex] :
                  dcrRow[colIndex];
            
            BrowseColumn<?> wcol = (BrowseColumn<?>) ThinClient.getInstance().getWidget(col.id);
            
            String display;
            try
            {
               if (!col.dataType.equalsIgnoreCase(var.getTypeName()) &&
                   var.getTypeName().equals("CHARACTER"))
               {
                  // change column type and format
                  
                  DisplayFormat fmt = getDisplayFormat(col);
                  
                  int len = fmt.getScreenWidth();
                  boolean rightAligned = fmt.isRightPositioning() ||
                                         (fmt instanceof NumberFormat);
                               
                  if (len <= 0)
                     len = 1;
                  
                  col.dataType = "CHARACTER";
                  col.format = "x(" + len + ")";
                  
                  if (rightAligned)
                     wcol.forceRightPositioning();
               }
               
               String txt = getDisplayFormat(col).fromVar(var).toScreenValue();
               
               display = renderCell(txt, col, colWidth, scroll);
            }
            catch (DisplayFormatCheckException ex)
            {
               display = DisplayFormat.generateQuestions(colWidth - 
                                                         Math.abs(scroll));
            }

            Color clr = col.columnDColor;
            screen().append(display, getColor(clr));
         }
      }
   }
  
}
