/*
** Module   : Browse.java
** Abstract : core browse widget implementation
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 SVG 20050803   @21995 Initial version.
** 002 SIY 20050831   @22486 Refactoring (see @@22394-22409). 
** 003 SIY 20050907   @22623 Fixed use of @inheritDoc tag in javadoc
** 004 SIY 20050909   @22819 Added attribute container, added and updated 
**                           methods for interaction with BrowseColumn, 
**                           added support for SENSITIVE attribute, 
**                           implemented Validateable interface 
**                           (temporarily commented out).
** 005 SIY 20051205   @24198 Added necessary methods to support Progress 
**                           methods and attributes.
** 006 SIY 20060131   @24199 Use ThinClient special method instead of
**                           direct calls to Window.message().
** 007 SIY 20060209   @24487 Updated to accommodate changes in xxxConfig
**                           classes.
** 008 SIY 20060315   @25119 Major changes to complete BROWSE 
**                           implementation.
** 009 SIY 20060324   @25274 Fixes to make behavior match Progress 
**                           counterpart.
** 010 NVS 20060703   @27766 Reviewed the uses of requestFocus() and
**                           setFocus(). They seem to be safe as they are
**                           internal focus changes.
** 011 NVS 20060703   @27799 getParent.nextFocus() replaced with
**                           ThinClient.nextFocus() call to maintain
**                           Progress semantics. Similar replacement has
**                           been done to getParent.previousFocus().
** 012 GES 20060713   @28001 Added support to honor the help attribute.
** 013 GES 20060808   @28493 Use int instead of long for row IDs. While
**                           this theoretically limits the maximum number 
**                           of rows per query, in practice the server
**                           side already has the same limitation AND it 
**                           is not practical to ever reach such limits
**                           in Progress code anyway. So we will take the
**                           small benefit of faster network performance
**                           over the capacity since the browse has a
**                           great potential to be a heavy consumer of
**                           network bandwidth.
** 014 SIY 20060824   @28844 Removed getP2JType() method.
** 015 SIY 20061025   @30637 Fixed size calculation and added support for
**                           column type override (@ fields).
** 016 SIY 20061027   @30757 Added refreshing of the browse on 
**                           VALUE-CHANGED event. Fixed drawing of empty
**                           browse.
** 017 SIY 20061102   @30939 Small change to encapsulate field.
** 018 SIY 20061105   @30982 Reposition browse to current row in the model.
** 019 SIY 20061114   @31361 Rewritten cache/navigation. Fixed insert mode
**                           handling and refreshing of row changes.
** 020 NVS 20061204   @31552 Switched to using a private widgetFocused
**                           flag instead of config.selected
** 021 ECF 20061204   @31579 Changed code to interpret a current row of -1
**                           to indicate a query off-end condition. A row
**                           should not be selected in this case.
** 022 ECF 20061206   @31588 Modified refreshStartRow(). This method now
**                           invalidates the browse instead of drawing it
**                           immediately.
** 023 ECF 20061207   @31621 Backed out #022 (@31588). This was causing a
**                           regression in that a reposition against the
**                           query was not correctly updating the browse
**                           view to reflect a new highlighted row.
** 024 SIY 20061208   @31628 refreshStartRow() now explicitly calls
**                           refresh().
** 025 SIY 20061208   @31654 Fix for initial viewport position and drawing
**                           of empty browse.
** 026 SIY 20061219   @31715 Adjusted refresh() behavior to make it similar
**                           to Progress.
** 027 SIY 20070315   @32418 Fixed page width calculation.
** 028 SIY 20070319   @32461 Set model current row in selectFocusedRow().
** 029 SIY 20070321   @32587 Fixed initialization of empty browse, fixes
**                           for layout calculations.
** 030 NVS 20070426   @33247 Enabled state is checked instead of SENSITIVE
**                           attribute.
** 031 SIY 20070504   @33420 New approach in model size handling.
** 032 SIY 20070515   @33747 Fixed positioning.
** 033 SIY 20070730   @34723 More positioning fixes.
** 034 SIY 20070731   @34743 Fixed handling of browse/column help messages.
** 035 SIY 20070926   @35251 Fixed numeric fields alignment. 
**                           Ref: #35250
** 036 ECF 20070928   @35272 Fixed row updates. Removed nullable check in
**                           stopEdit(). Added internal, END-ERROR driven,
**                           row level undo processing to match Progress'
**                           behavior on failed field constraint checks.
** 037 ECF 20071011   @35447 Key handling fixes. Added handler for END
**                           function (move to last row). Fixed PAGE UP
**                           and PAGE DOWN handlers. Disallow negative row
**                           index in setCurrentRow(), except for special
**                           case of END handler.
** 038 ECF 20071030   @35631 Modified refresh(boolean). Added code to draw
**                           after refreshing the data cache. Modified
**                           refreshStartRow() to remove draw() call,
**                           since it is now redundant.
** 039 GES 20071227   @36593 Added hasFocusable() method.
** 040 SIY 20080121   @36873 Changed sending of VALUE-CHANGED. Attempt  
**                           to limit redrawing in some cases.
**                           Ref: #36871
** 041 SIY 20080128   @36920 More correct checking of conditions when 
**                           redraw is required.
**                           Ref: #36919
** 042 SVL 20080208   @36972 Force right-aligning when the column format
**                           is changed from 'number' to 'character'.
** 043 NVS 20080401   @37770 Added isFocusable() method implementation.
** 044 SIY 20080409   @37883 Fixed focus and keyboard and focus handling.
**                           Ref: #37887
** 045 NVS 20080414   @37982 ENTRY event is never consumed by Browse and
**                           may only be consumed by NO-APPLY triggers.
** 046 SIY 20080416   @37995 Postponed refreshing if frame is not visible.
**                           Ref: #37994
** 047 SIY 20080505   @38213 Fixed drawing of not focused BROWSE.
** 048 SIY 20080506   @38217 Force reinitialization of the underlying query
**                           when it is known that previously query
**                           returned other values. Some performance
**                           improvements.
**                           Ref: #38216
** 049 SIY 20080512   @38255 Fixed drawing of disabled widget.
**                           Removed #048, persistence fixes made it 
**                           redundant. More performance improvements.
** 050 SIY 20080521   @38362 Fixed keyboard handling for editing mode.
**                           Ref: #38355
** 051 SIY 20080617   @38796 Fixed initial display of empty browse.
** 052 SIY 20080623   @38925 Fixed drawing of special case when browse is
**                           displayed before query is assigned.
** 053 SIY 20080703   @39036 Fixed drawing of horizontal delimiter.
** 054 CA  20080901   @39644 Fixed startEdit: must show the help for the
**                           the focused widget.
** 055 SIY 20080903   @39695 More fixes in drawing of delimiters and
**                           highlighting.
** 056 CA  20080911   @39784 Bug fix in DataColumnRenderer.addData - the
**                           column and var type must be compared using a
**                           case-insensitive method.
** 057 SVL 20080925   @39941 Call to repaint() was added to
**                           refreshStartRow().
** 058 SIY 20081105   @40313 Added flag which tracks 'focus attractor'
**                           state. See #40284 for more details.
** 059 SIY 20081107   @40368 Rolled back #058, it is not necessary since
**                           more generic approach is discovered.
**                           See #40367 for more details.
** 060 CA  20081114   @40451 Changes due to new clipping implementation.
** 061 CA  20081203   @40757 When an event is received, clear the temp
**                           texts from the message area.
** 062 CA  20081208   @40821 Added ROW-DISPLAY event implementation.
** 063 CA  20081210   @40845 Implemented CLOSE QUERY statement - on close,
**                           the browse widget will be notified to reset
**                           its state.
** 064 CA  20081211   @40861 Fixed a regression introduced by H062: after
**                           ROW-DISPLAY trigger invocation, reposition 
**                           the query to the previous 'scrollRow' value.
** 065 CA  20081212   @40884 Yet another regression about ROW-DISPLAY: the
**                           current row must be computed from scrollRow 
**                           and currRow.
** 066 CA  20081212   @40887 On CLOSE QUERY, reset the currRow field.
**                           triggerRowDisplayEvents: row is restored from
**                           the currRow value.
** 067 SIY 20090213   @41309 Added 'force' parameter into refreshStartRow
**                           signature.
** 068 NVS 20090213   @41312 Name changed for clearPutScreenInMessageArea
**                           method in ThinCLient.
** 069 SIY 20090216   @41322 Added tracking of highlighting state.
** 070 CA  20090403   @41728 Fix regression introduced by frame disabling
**                           on destroy - generate ROW-DISPLAY events only
**                           if frame is still alive.
** 071 SIY 20090413   @41784 Fix regression caused by changes in focus 
**                           management. Do not remove highlighting on 
**                           CHARVA focus change since it does not reflect
**                           P2J widget highlighting.
** 072 CA  20090423   @41926 Fixed a regression introduced by ROW-DISPLAY
**                           events - these events must be triggered only
**                           on initial draw or when the rows have scrolled.
**                           The problem was that, on each draw, the query was 
**                           repositioned (which might interfere with an
**                           ongoing CREATE/UPDATE operation).
** 073 SIY 20090902   @43824 Browse is always valid (just like other simple 
**                           components).
** 074 SIY 20090908   @43840 Do not redraw widget if parent frame is hidden.
** 075 SVL 20091011   @44133 Now we take into account the offset of the zero
**                           row of the rowDisplayed array (which equals
**                           dcStartRow).
** 076 SVL 20091029   @44287 Reworked logic into refresh(boolean) responsible
**                           for determination of the top row in a view.
** 077 SVL 20091105   @44315 Into refresh(boolean) do not set the current row
**                           as the start row in a view if the total number of
**                           rows is less than or equals view size.
** 078 SVL 20091119   @44441 Fixed repositioning for editable browse, fixed
**                           adjustment of current row on focus gaining event
**                           if the current row of the backing query row was
**                           changed because of GET statement, fixed
**                           browse:refresh() for the case of server-side
**                           update of rows.
** 079 EVL 20100430   @44839 Fix for incorrect help string with empty browser
**                           data.
** 080 SIY 20100619          Switching to own implementation of character UI.
** 081 SIY 20100804          Changes in imports, some refactorings.
** 082 SIY 20100822          Refactored AttributeContainer support.
** 083 SIY 20100015          locateFrame() is moved to UiUtils, renamed base
**                           class.
** 084 SIY 20101122          Switching from KeyEvent to KeyInput.
** 085 SIY 20101202          Minor change in constant name.
** 086 SIY 20110112          Changes related to new color handling scheme.
** 087 SIY 20111101          Updated imports, minor cleanups/refactorings.
** 088 SIY 20111202          Refactored and moved from CHUI package.
** 089 SIY 20111221          Switched to widget factory.
** 090 CA  20121022          Added support for optional query and column-list clauses.
** 091 GES 20140627          Changed color attribute method names to match naming standards.
** 092 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access 
**                           only to the base APIs provided by the OutputManager.
** 093 HC  20140911          GUI coordinate support - data types refactoring.
** 094 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 095 VIG 20140923          Changed clearPutScreenInMessageArea and message calls to window-aware.
** 096 CA  20141104          Replaced BrowseConfig.width with BaseConfig.widthChars.
** 097 HC  20141021          Implemented window move/position support.
** 098 SVL 20141209          Added support for NO-UNDERLINE. Partially fixed refreshing. Fixed
**                           row selection. Implemented some new functions. Added support for
**                           locked columns.
** 099 HC  20141207          Implemented runtime support for widget coordinate attributes.
** 100 SVL 20150212          Browse reworked.
** 101 SVL 20150305          Fixed SELECT-ROW for editing browse. Added refresh mode for
**                           DELETE-CURRENT-ROW.
** 102 SVL 20150331          Implemented selected row availability mechanics (required for
**                           DELETE-* functions). Implemented ROW-LEAVE triggers.
** 103 SVL 20150413          Implemented SELECT-{NEXT|PREV}-ROW.
** 104 HC  20150323          Removed business logic from config classes.
** 105 SVL 20150420          Implemented REFRESHABLE and ADD-LIKE-COLUMN.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Implementation of Progress BROWSE widget.
 * 
 * TODO:  Do we handle options like NO-ASSIGN, NO-VALIDATE, and VALIDATE
 * functions?
 */
public abstract class Browse<O extends OutputManager<?>> 
extends FixedSizeContainer<O>
implements WidgetWithConfig<BrowseConfig>,
           FocusListener,
           KeyListener
{
   /** The size of data cache in screen pages. */
   public static final int CACHE_SIZE = 20;
   
   /** A data model instance. */
   protected BrowseModel browseModel = null;

   /**
    * Is in-place edit can be enabled. This is indicated with '*' sign in
    * the leftmost column.
    */
   protected boolean editPossible = false;
   
   /** Determines if the current row is selected (single-selection mode only). */
   protected boolean currRowSelected = true;
   
   /**
    * Index of current column. It is the column where the last edit is
    * performed.
    */
   protected int currColumn = 0;
   
   /**
    * An updated data for the current row. These data are passed to model if
    * focus leaves this row. If the array or some element in it is
    * <code>null</code> then no updates are available for the corresponding
    * cell.
    */
   protected BaseDataType[] currRowUpdates = null;
   
   /**
    * Indexes of selected rows in <code>Integer</code> form. Used only in
    * multiple selection mode. This mode is active if and only if this member
    * is not <code>null</code>.
    */
   protected LinkedHashSet<SelectedRow> selectedRowIndices = null;
   
   /**
    * The data model index of the scroll anchor column. Value <code>0</code>
    * means the absence of horizontal scrolling. Otherwise this column will
    * occupy the right part of the view window.
    */
   protected int scrollColumn = 0;
   
   /**
    * The height (in character rows) of the labels area not including the
    * separator line.
    */
   protected int ilLabelsHeight = 0;
   
   /** Number of character character columns in the displayable page. */
   protected double ilPageWidth = 0;
   
   /** Number of rows in the displayable page. */
   protected int ilPageRowsCount = 0;
   
   /**
    * Number of space columns on screen between the data area and the right
    * border.
    */
   protected double ilRightSpaceWidth = 0;
   
   /**
    * The minimal positive value of <code>scrollColumn</code>.
    */
   protected int ilFirstScrollCol = 0;
   
   /**
    * Cached rows from the data model. Cache size matches the number of rows displayed in the
    * viewport.
    */
   protected BaseDataType[][] dcData = null;
   
   /** Separator line length */
   protected double separatorLen = 0;
   
   /**
    * Rendered columns labels area. This cache is reset when horizontal
    * scrolling is performed.
    */
   protected String[] rcLabels = null;
   
   /** Component attribute container. */
   protected BrowseConfig config = null;
   
   /** Draw labels. */
   protected boolean drawLabels = true;

   /** 0-based index of the current row in the browse model. */
   private int currentRow = 0;

   /**
    * Determines if the current row was available (could be fetched) at the moment it has become
    * the current row.
    */
   private boolean currentRowAvail = false;

   /** Help message for current column. */
   private String columnHelp = null;
   
   /** Waiting for END-ERROR key after edit mode validation error */
   private boolean awaitingEndError = false;

   /** Set to true if browse was drawn at least once. */
   protected boolean hadDraw = false;
   
   /** 
    * This variable is used to reproduce special drawing behavior for case
    * when query is initialized after browse is displayed first time. 
    */
   protected boolean pseudoFocus = false;

   /**
    * The number that allows to assess if the total number or rows is more or less than the page
    * size. Can take the following values:
    * <ul>
    *    <li><code>-1</code> if the query is closed or we haven't got any rows yet;</li>
    *    <li>less or equal page size if that represents the number of rows;</li>
    *    <li><code>page size + 1</code> if the number of rows is more than the page size.</li>
    * </ul>
    */
   protected int rowCountAssessment = -1;

   /** Set to <code>true</code> when separators need to be underscored. */
   public boolean separatorQuirk = false;

   /** Keep track of each row display status. */
   private boolean[] rowDisplayed = null;

   /** The P2J DCOLOR for each cell. */
   protected Color[][] colors = null;

   /** Protect ROW-DISPLAY events against recursion. */
   private boolean rowDisplayTriggered = false;

   /**
    * Width of browse as it was passed to {@link #setSize(double, double)}. Used to recalculate
    * size variables when a new column is added.
    */
   private double originalWidth;

   /**
    * Height of browse as it was passed to {@link #setSize(double, double)}. Used to recalculate
    * size variables when a new column is added.
    */
   private double originalHeight;

   /** Info about at which row started the last drawing. */
   protected int oldDisplayRow = -1;

   /**
    * Terminal color for current text. It is displayed in inverse colors.
    */
   public Color colorCurrent;

   /**
    * Terminal color for normal text in data area when separators are
    * active. It is displayed with underline.
    */
   protected Color colorHSep;

   /** Terminal attribute for normal text. */
   public Color colorCommon;

   /**
    * Constructs the Browse widget.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public Browse(WidgetId id, BrowseConfig cfg)
   {
   }

   /**
    * Get the index of the row on the top of the page.
    *
    * @return 0-based index (in the browse model) of the row on the top of the page or
    *         <code>0</code> if there are no rows.
    */
   protected int getScrollRow()
   {
      if (dcData == null || dcData.length == 0)
         return 0;

      return getCachedRowNumber(0);
   }

   /**
    * Get the current row.
    *
    * @return 0-based index of the current row in the browse model.
    */
   protected int getCurrentRow()
   {
      return currentRow;
   }

   /**
    * Get the current row as {@link com.goldencode.p2j.ui.client.Browse.SelectedRow} object which
    * takes into account its availability (see {@link #currentRowAvail}).
    *
    * @return see above.
    */
   protected SelectedRow getCurrentRowAsSelected()
   {
      return new SelectedRow(currentRow, currentRowAvail);
   }

   /**
    * Get the browse model index of the data row positioned at the specified viewport row.
    *
    * @param  viewportIndex
    *         0-based index of the target row in the viewport.
    *
    * @return 0-based index of the target row in the browse model or <code>-1</code> if there are
    *         no rows or there is no data row at the specified viewport row.
    */
   protected int getCachedRowNumber(int viewportIndex)
   {
      if (dcData == null || viewportIndex > dcData.length - 1)
         return -1;

      return ((integer) dcData[viewportIndex][0]).intValue();
   }

   /**
    * Get the number of rows in the cache (which matches the number of rows in the viewport).
    *
    * @return the number of rows in the cache.
    */
   protected int getCacheSize()
   {
      return dcData == null ? 0 : dcData.length;
   }

   /**
    * Get the index of the current row in the viewport.
    *
    * @return 0-based index of the current row in the viewport.
    */
   protected int getViewportRow()
   {
      return getViewportRow(getCurrentRow());
   }

   /**
    * Get the index of the specified row in the viewport.
    *
    * @param  rowIndex
    *         0-based index of the target row in the browse model.
    *
    * @return 0-based index of the specified row in the viewport. <code>0</code> if the row out
    *         of the current view.
    */
   protected int getViewportRow(int rowIndex)
   {
      if (dcData == null)
         return 0;

      for (int i = 0; i < dcData.length; i++)
      {
         if (getCachedRowNumber(i) == rowIndex)
            return i;
      }

      return 0;
   }

   /**
    * Scroll browse by the specified number of rows. If the number of rows is less than the page
    * size, then some rows from the current view remain in the cache (without refresh). The
    * new rows are obtained from the server starting from the first row which row index in the
    * browse model is more than the index of the last row in the viewport (or less than the index
    * of the first row if we scroll in backward direction).
    *
    * @param  rowCount
    *         Number of rows to scroll. Positive count to scroll forward, negative to scroll
    *         backward.
    *
    * @return <code>true</code> if at least one row was obtained from the server.
    */
   protected boolean incrementalScroll(int rowCount)
   {
      if (rowCount == 0)
         return false;

      if (dcData.length != ilPageRowsCount && rowCount > 0)
         throw new IllegalStateException("There is no point to use incrementall scroll for " +
                                         "scrolling forward if the page is not full.");

      boolean forward = rowCount > 0;
      int diff = Math.abs(rowCount);
      int overlap = Math.min(ilPageRowsCount - diff, getCacheSize());

      BrowseColumnConfig[] columns = browseModel.getColumns();
      int colNo = columns.length;

      int startRow = forward ? getCachedRowNumber(ilPageRowsCount - 1) + 1 :
                               getCachedRowNumber(0) - 1;

      BaseDataType[][] serverData = browseModel.getRows(startRow, diff, forward);
      if (serverData.length > 0)
      {
         Color[][] newColors = new Color[overlap + serverData.length][colNo];
         if (overlap > 0)
         {
            BaseDataType[][] newData = new BaseDataType[overlap + serverData.length][colNo];

            rowDisplayed = new boolean[newData.length];
            Arrays.fill(rowDisplayed, false);

            if (forward)
            {
               System.arraycopy(dcData, getCacheSize() - overlap, newData, 0, overlap);
               System.arraycopy(colors, getCacheSize() - overlap, newColors, 0, overlap);

               Arrays.fill(rowDisplayed, 0, overlap, true);

               System.arraycopy(serverData, 0, newData, overlap, serverData.length);
            }
            else
            {
               System.arraycopy(serverData, 0, newData, 0, serverData.length);

               System.arraycopy(dcData, 0, newData, serverData.length, overlap);
               System.arraycopy(colors, 0, newColors, serverData.length, overlap);
               Arrays.fill(rowDisplayed, serverData.length, serverData.length + overlap, true);
            }
            dcData = newData;
         }
         else
         {
            dcData = serverData;
            rowDisplayed = new boolean[serverData.length];
            Arrays.fill(rowDisplayed, false);
         }

         rowDisplayTriggered = false;
         colors = newColors;
      }

      return serverData.length > 0;
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, BrowseConfig cfg)
   {
      config = ConfigManager.getInstance().resolveWidgetConfig(id, cfg);
      ConfigManager.getInstance().replaceConfig(config, cfg);
      
      addKeyListener(this);
      addFocusListener(this);
   }

   /**
    * Destroy this widget.  Notify all of its components that the widget has been destroyed.
    */
   @Override
   public void destroy()
   {
      super.destroy();

      if (browseModel == null)
      {
         return;
      }
      
      BrowseColumnConfig[] cols = browseModel.getColumns();
      ThinClient tc = ThinClient.getInstance();
      
      for (BrowseColumnConfig col : cols)
      {
         @SuppressWarnings("unchecked")
         BrowseColumn<O> bcol = (BrowseColumn<O>) tc.getWidget(col.id);

         if (bcol != null)
         {
            bcol.destroy();
         }
      }
   }

   /**
    * Get display format which corresponds to format string and data type
    * stored in configuration.
    * 
    * @return  Instance of <code>DisplayFormat</code>.
    */
   public static DisplayFormat getDisplayFormat(BrowseColumnConfig colCfg)
   {
      DisplayFormat colFormat = null;
      
      try
      {
         BaseDataType var = DisplayFormat.instanceOfType(colCfg.dataType);
         
         if (colCfg.format != null)
            colFormat = DisplayFormatFactory.makeFormat(var, colCfg.format);
         else
            colFormat = DisplayFormatFactory.makeDefaultFormat(var);
      }
      catch (DisplayFormatParsingException e)
      {
         throw new RuntimeException(e);
      }
      
      return colFormat;
   }

   /**
    * Return label as array of strings.
    * 
    * @return  Parsed label.
    */
   public static String[] getLabelVar(BrowseColumnConfig colCfg)
   {
      String label = colCfg.clabel;
      
      if (label == null)
         label = colCfg.label;
      
      if (label == null || (!colCfg.labels && !colCfg.forceLabel))
         return new String[0]; // no label
   
      int len = label.length();
      boolean singlelined = true;
   
      int offsExcl = 0;
      while (offsExcl < len)
      {
         offsExcl = label.indexOf('!', offsExcl);
         if (offsExcl < 0)
         {
            break; // no single '!'
         }
   
         ++offsExcl;
         if (offsExcl == len)
         {
            break; // no char after single '!'
         }
   
         if (label.charAt(offsExcl) == '!')
         {
            ++offsExcl;
            continue; // skipping double '!'
         }
   
         // found single '!' followed by other char. Now checking two last
         // chars. If both are '!' then label is contains only one line.
         if (label.charAt(len - 1) != '!' || label.charAt(len - 2) != '!')
         {
            singlelined = false;
         }
         break;
      }
   
      if (singlelined)
         return new String[] { label }; // displaying label as-is
   
      ArrayList<String> lines = new ArrayList<String>(4);
      StringBuilder line = new StringBuilder(80);
   
      int offsLine = 0;
      while (true)
      {
         int offsSep = label.indexOf('!', offsLine);
         
         if (offsSep < 0)
         {
            offsSep = len;
         }
         line.append(label.substring(offsLine, offsSep));
   
         // last line is indicated as no '!' or no chars after next '!'
         if (offsSep >= len - 1)
         {
            // finishing last line
            lines.add(line.toString());
            break;
         }
   
         ++offsSep;
         if (label.charAt(offsSep) == '!')
         {
            // double '!' should be treated as one '!' in screen value
            line.append('!');
            ++offsSep;
         }
         else
         {
            // real line separator
            lines.add(line.toString());
            line.setLength(0);
         }
   
         offsLine = offsSep;
      }
   
      return lines.toArray(new String[lines.size()]);
   }

   /**
    * @see com.goldencode.p2j.ui.client.widget.Widget#setEnabled(boolean)
    */
   @Override
   public void setEnabled(boolean enabled)
   {
      if (enabled && hadDraw)
         separatorQuirk  = true;

      if(enabled != isEnabled())
         pseudoFocus = false;

      if (enabled)
      {
         resetRowDisplayState();
      }

      super.setEnabled(enabled);
   }
   
   /**
    * Tracks the highlighted state  of the component.
    * 
    * @param  highlighted
    *         New highlighted state.
    */
   @Override
   public void setHighlighted(boolean highlighted)
   {
      if (highlighted != highlighted())
         pseudoFocus = false;
      
      super.setHighlighted(highlighted);
   }
   
   /**
    * Second stage of the initialization of the Browse widget which is
    * performed only when query is assigned.
    * 
    * @param   queryId
    *          Server side query ID.
    */
   public void setQuery(int queryId)
   {
      if (browseModel != null)
         return;

      browseModel = new ClientBrowseModel(queryId);

      BrowseColumnConfig[] cols = browseModel.getColumns();

      ilLabelsHeight = 0;
      ilPageRowsCount = config.down;
      
      int minHeight = config.down;
      int minWidth  = 4;

      if (config.box)
      {
         minWidth += 2;
         minHeight += 2;
      }
      
      editPossible = false;
      int firstEditable = -1;
      
      drawLabels = config.labels;
      
      for (int i = 0; i < cols.length; ++i)
      {
         if (!config.labels)
         {
            // no need to set the column's LABELS attr: this is set right by the server-side
            if (cols[i].forceLabel)
               drawLabels = true;
         }
         
         if (!cols[i].readOnly)
         {
            editPossible = true;
            
            if (firstEditable == -1)
               firstEditable = i;
         }
         
         minWidth += cols[i].widthChars;
         
         if (i < (cols.length - 1))
            minWidth += 1;
      }

      updateLabelsHeigth();
      
      if (drawLabels)
         minHeight += ilLabelsHeight + (config.noUnderline ? 0 : 1); //Headers
      
      if (config.widthChars > 0)
         minWidth = (int) Coordinate.scale(config.widthChars);
      
      if (config.widthChars == 0)
      {
         config.widthChars = Coordinate.scale(minWidth); // direct access
      }
      if (config.heightChars == 0)
      {
         config.heightChars = Coordinate.scale(minHeight); // direct access
      }
      
      setSize(config.widthChars, config.heightChars);
      
      if (config.multiple)
         selectedRowIndices = new LinkedHashSet<>();

      // init cache
      rowCountAssessment = -1;
      scrollToRow(0, true, true);

      if (editPossible && getCacheSize() > 0)
      {
         if (selectedRowIndices != null)
         {
            selectedRowIndices.add(new SelectedRow(0, false));
         }
         currColumn = firstEditable;
         
         startEdit();
      }
      else if (editPossible && firstEditable >= 0)
         config.help = cols[firstEditable].help;

      if (config.multiple)
         setSelectionMultiple(true);
   }

   /**
    * Getting the BaseDataType object representing the current value.
    *
    * @return  The value representing the current widget value.
    */
   public BaseDataType getValue()
   {
      // editing is allowed
      if (getActiveEditor() != null)
         return getActiveEditor().getValue(); 
         
      // editing is not allowed
      // TODO: change the approach
      // Code below may cause problems in GUI environment, because each 
      // validation will reset view to selected cell. This is not an issue
      // in TUI because there is no way to make current cell not in view port. 
      columnToView(currColumn);
      
      return getValue(currColumn);
   }
   
   /**
    * Check if value is modified since last commit.
    * 
    * @return  <code>true</code> if value was changed.
    */
   public boolean isModified()
   {
      return isCurrentRowModified() || config.modified;
   }

   /**
    * Handle situation when <code>Browse</code> widget receives the focus.
    * <p>
    * When a help string is defined for the browse it will be displayed on
    * the status line.  This method also sets the selected attribute on. 
    * 
    * @param    event
    *           Reference to the event.
    */
   @Override
   public void onFocusGained(FocusEvent event)
   {
      // For editable browse it is actually a "focus lost" event, do not update selection
      // in this case.
      boolean updateSelection = getActiveEditor() == null;

      setCurrentRow(getCurrentRow(), updateSelection, true);

      if (columnHelp != null && config.help == null)
      {
         columnHelp = null;
      }
      else
      {
         ThinClient.setHelp(config.help);
      }

      repaint();
   }
   
   /**
    * Handle situation when <code>Browse</code> widget loses the focus.
    * <p>
    * This method clears the focused flag.
    * 
    * @param    fe_
    *           Reference to the event.
    */
   @Override
   public void onFocusLost(FocusEvent fe_)
   {
      repaint();
   }
   
   /**
    * Undo editing operation results and reset everything to initial state.
    */
   public void undoValue()
   {
      FillIn<O,?> editor = getActiveEditor();
      
      if (editor != null)
         stopEdit();
      
      ConfigHelper.setModified(config, false);
      currRowUpdates = null;
      
      if (editor != null)
         startEdit();
   }
   
   /**
    * Browse behaves like simple component in this regard.
    * 
    * @return  <code>true</code>.
    */
   public boolean valid()
   {
      return true;
   }

   /**
    * Requesting the focus to the <code>Browse</code>.
    */
   @Override
   public void requestFocus()
   {
      if (!isEnabled())
         return;

      setCurrentRow(getCurrentRow(), true, true);

      if (getActiveEditor() != null && selectedRowIndices != null)
      {
         selectedRowIndices.clear();
         selectedRowIndices.add(getCurrentRowAsSelected());
         draw();
         requestSync();
      }

      super.requestFocus();
   }
   
   /**
    * Set READ-ONLY attribute.
    * 
    * @param   readOnly
    *          New value for the attribute.
    */
   public void setReadOnly(boolean readOnly)
   {
      // Config may be overwritten in unexpected places (e.g. in getRows because of maxDataGuess
      // update). So we want to avoid calling startEdit if we have temporarily stopped editing
      // in order to update cache contents.
      if (config.readOnly == readOnly)
         return;

      config.readOnly = readOnly;

      if (browseModel == null)
         return;

      if (readOnly)
         stopEdit();
      else
         startEdit();
   }

   /**
    * Select all rows.
    */
   public void selectAll()
   {
      if (isSelectionMultiple())
      {
         boolean editing = getActiveEditor() != null;
         if (editing)
         {
            undoValue();
            stopEdit();
         }

         // get the last row
         int bigNumber = Integer.MAX_VALUE - 1000; // decreased to avoid overflow
         BaseDataType[][] row = browseModel.getRows(bigNumber, 1, false);

         selectedRowIndices.clear();
         if (row.length == 0)
            return;

         int numResults = ((integer) row[0][0]).intValue() + 1;
         for (int i = 0; i < numResults; i++)
         {
            SelectedRow selectedRow = new SelectedRow(i);
            selectedRowIndices.add(selectedRow);
         }

         // scroll to the first row in the result set
         boolean anyRows = scrollToRow(0, false, true);
         if (anyRows)
            setCurrentRow(getCachedRowNumber(0));

         if (editing)
         {
            startEdit(false);
         }
      }
   }
   
   /**
    * Remove all selections. 
    */
   public void deselectAll()
   {
      if (selectedRowIndices != null)
         selectedRowIndices.clear();
      else
         currRowSelected = false;

      draw();
      requestSync();
   }
   
   /**
    * Select row which is currently in focus. 
    */
   public void selectFocusedRow()
   {
      if (getCurrentRow() < 0 || browseModel == null)
      {
         return;
      }
      
      if (selectedRowIndices != null)
      {
         SelectedRow currRow = getCurrentRowAsSelected();
         if (!selectedRowIndices.contains(currRow))
         {
            selectedRowIndices.add(currRow);
            draw();
            requestSync();
         }
      }
   }

   /**
    * Implements the SELECT-ROW() widget method.<p>
    * Selects the specified row if it is currently in the browse viewport. In a single-select
    * browse, the previously selected row is deselected. No rows are deselected in a
    * multiple-select browse.
    *
    * @param   rowIndex
    *          0-based row index in the browse viewport.
    *
    * @return  <code>true</code> if the row with the given index exist in the browse viewport.
    */
   public boolean selectRow(int rowIndex)
   {
      if (rowIndex < 0 || rowIndex > getCacheSize() - 1)
      {
         return false;
      }

      int row = getCachedRowNumber(rowIndex);
      SelectedRow selectedRow = new SelectedRow(row);
      if (selectedRowIndices != null)
      {
         if (!selectedRowIndices.contains(selectedRow))
         {
            boolean avail = browseModel.isRowAvailable(row);
            selectedRowIndices.add(new SelectedRow(selectedRow.getRowIndex(), avail));
            draw();
            requestSync();
         }
      }
      else if (getCurrentRow() != row)
      {
         boolean editing = getActiveEditor() != null;
         if (editing)
         {
            releaseCurrentRow(false);
         }

         setCurrentRow(row);
         draw();
         requestSync();

         if (editing)
            startEdit();
      }

      return true;
   }

   /**
    * Deselect row which is currently in focus.
    */
   public void deselectFocusedRow()
   {
      if (getCurrentRow() < 0 || browseModel == null)
      {
         return;
      }

      if (selectedRowIndices != null)
      {
         SelectedRow currRow = getCurrentRowAsSelected();
         if (selectedRowIndices.contains(currRow))
         {
            selectedRowIndices.remove(currRow);
            draw();
            requestSync();
         }
      }
      else
      {
         currRowSelected = false;
         draw();
         requestSync();
      }
   }

   /**
    * Implements the DESELECT-SELECTED-ROW() widget method.<p> Deselects the <i>n</i>th selected
    * row in a browse.<p>
    * Progress maintains a numbered list of selected rows, starting at 1. When the
    * DESELECT-SELECTED-ROW(n) method is encountered, Progress searches this list to find
    * the <i>n</i>th selected row.
    *
    * @param   rowIndex
    *          0-based index of the row (in the list of selected rows) to deselect.
    *
    * @return  <code>true</code> if the row is successfully deselected.
    */
   public boolean deselectSelectedRow(int rowIndex)
   {
      if (rowIndex < 0)
         return false;

      if (selectedRowIndices != null)
      {
         if (rowIndex > selectedRowIndices.size() - 1)
            return false;

         // skip elements
         Iterator iter = selectedRowIndices.iterator();
         for (int i = 0; i < rowIndex; i++)
         {
            iter.next();
         }

         // remove target element
         iter.next();
         iter.remove();

      }
      else
      {
         if (rowIndex > 0 || !currRowSelected)
            return false;

         if (hasFocus() && isEnabled() && (getActiveEditor() == null))
         {
            // Progress quirk: when a non-editable browse has focus, the row is NOT deselected,
            // but true is returned.
            return true;
         }

         currRowSelected = false;
      }

      draw();
      requestSync();

      return true;
   }
   
   /**
    * Renders the screen value into a cell. This is helper method for the
    * render classes.
    * 
    * @param   screen
    *          Screen value.
    * @param   col
    *          Column to render.
    * @param   colWidth
    *          The full width for formatting.
    * @param   scroll
    *          Scrolling correction. Means the number of characters which
    *          should be erased from the left (if positive) or from the right
    *          (if negative). This scroll differs from the scroll in
    *          <code>DisplayFormat.formatScreenValue()</code> because it
    *          means the scroll of the whole browse column not the scroll of
    *          screen value inside the column.
    * 
    * @return  Rendered screen value.
    */
   protected static String renderCell(String screen, BrowseColumnConfig col, 
                                     int colWidth, int scroll)
   {
      BrowseColumn<?> colw = (BrowseColumn<?>) ThinClient.getInstance().getWidget(col.id);
      boolean rightAlign = getDisplayFormat(col).isRightPositioning() || colw.isRightPositioning();
      
      // adjust alignment for numbers
      if (getDisplayFormat(col) instanceof NumberFormat)
         rightAlign = true;
      
      if (colw.isRightPositioning())
         screen = screen.trim();
      
      String display = DisplayFormat.formatScreenValue(screen, colWidth, 0, 
                                                       rightAlign, rightAlign);
      if (scroll > 0)
      {
         // scroll by truncating from the left
         display = display.substring(scroll);
      }
      else
      {
         display = display.substring(0, display.length() + scroll);
      }

      return display;
   }

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @Override
   public BrowseConfig config()
   {
      return config;
   }
   
   /**
    * Sets position of terminal's cursor. Redirects this positioning to the
    * in-place editor if it is present. Otherwise does nothing because BROWSE
    * itself has no purpose for the hardware cursor.
    */
   @Override
   public abstract void drawCaret();

   /**
    * Draws the widget: border, header, selection marks and line, internal
    * FillIn etc.
    */
   @Override
   public abstract void draw();

   /**
    * Sets new widget's size.
    *
    * @param width
    *        The width component of the new size.
    * @param height
    *        The height component of the new size.
    *
    * @throws IllegalArgumentException
    *         When new size is not large enough.
    */
   @Override
   public void setSize(double width, double height)
   {
      originalWidth = width;
      originalHeight = height;

      width = Coordinate.scale(width);
      height = Coordinate.scale(height);
      
      if (browseModel == null || width == 0 || height == 0)
         return;

      // mark columns
      ilPageWidth = width - 2;
      ilPageRowsCount = (int) height;

      BrowseColumnConfig[] cols = browseModel.getColumns();
      
      double colWidth = 0;
      
      for (int i = 0; i < cols.length; i++)
      {
         colWidth += cols[i].widthChars + 1;
      }
      
      colWidth = Coordinate.scale(colWidth);
      
      if (cols.length > 0)
      {
         // the last column has no vertical separator
         colWidth--;
      }
      
      if (config.box)
      {
         // if no columns specified, use the explicit width set by the user.
         // safe, comparison of scaled doubles
         ilPageWidth = cols.length > 0 
                          ? Math.min(colWidth, width - 4)
                          : width - 4;
         ilPageRowsCount -= 2;
      }

      if (drawLabels)
      {
         ilPageRowsCount -= ilLabelsHeight + (config.noUnderline ? 0 : 1);
      }

      // safe, comparison of scaled doubles
      if (ilPageWidth < 1 || ilPageRowsCount < 2)
      {
         throw new IllegalArgumentException
            ("Too little browse size: " + width + "x" + height +
             ". It implies the data area size: " + ilPageWidth + "x" +
             ilPageRowsCount);
      }

      ilRightSpaceWidth = Coordinate.scale(ilPageWidth - colWidth);
      
      // safe, comparison of scaled doubles
      if (ilRightSpaceWidth <= 0)
      {
         ilRightSpaceWidth = 0;
      }
      
      ilFirstScrollCol = cols.length;
      double w = -1;
      int availableWidth = getAvailableWidth();
      
      for (int i = 0; i < cols.length; ++i)
      {
         w += cols[i].widthChars + 1;
         
         // safe, comparison of scaled doubles
         if ((w = Coordinate.scale(w)) > ilPageWidth)
         {
            if (cols[i].widthChars > availableWidth)
            {
               // there is not enough space to entirely display the first scroll column, so
               // the next column becomes the first scroll column
               ilFirstScrollCol = i + 1;
            }
            else
            {
               ilFirstScrollCol = i;
            }
            break;
         }
      }

      separatorLen = ilPageWidth - (config.box ? 0 : 2);

      super.setSize(width, height);
   }

   /**
    * This component will receive an input focus when Tab or Shift-Tab
    * is pressed.
    * 
    * @return <code>true</code> if this component is enabled and visible.
    */
   @Override
   public boolean focusTraversable()
   {
      return isEnabled() && isVisible();
   }

   /**
    * Returns the same as <code>getDefaultFocus()</code> because focus placement
    * inside BROWSE can't be switched from initial component until subcomponent
    * creation or deletion.
    * 
    * @see com.goldencode.p2j.ui.client.widget.Container#currentFocus()
    */
   @Override
   public Widget<O> currentFocus()
   {
      return defaultFocus();
   }

   /**
    * Returns this component directly if there is no in-place editor.
    * Otherwise returns this editor.
    * 
    * @see com.goldencode.p2j.ui.client.widget.Container#defaultFocus()
    */
   @Override
   public Widget<O> defaultFocus()
   {
      FillIn<O,?> editor = getActiveEditor();
      if (editor !=null)
      {
         return editor;
      }

      return this;
   }
                                         
   /**                                             
    * Reports if there is at least 1 focusable column that is not the
    * current column.
    *
    * @param    next
    *           <code>true</code> if this must search for a next focusable
    *           widget, <code>false</code> for a previous focusable widget.
    * @param    exclude
    *           This is ignored since the current column is internally known
    *           by the browse.
    *
    * @return   <code>true</code> if there is at least one other column
    *           that can be focused.
    */
   public boolean hasFocusable(boolean next, Widget<O> exclude)
   {
      // quick exit if no edits are possible
      if (!editPossible)
         return false;
      
      // when editing is not active, browse acts like a single widget that
      // has no focusable widgets inside
      if (getActiveEditor() == null)
         return false;
      
      BrowseColumnConfig[] cols = browseModel.getColumns();
      
      // can't previous past beginning of first row
      if (!next && getCurrentRow() == 0)
      {
         for (int j = currColumn - 1; j >= 0; j--)
         {
            if (!cols[j].readOnly)
               return true;
         }
         
         return false;
      }
      
      // can't next past end of last row
      if (next && getCurrentRow() == (getCacheSize() - 1))
      {
         for (int k = currColumn + 1; k < cols.length; k++)
         {
            if (!cols[k].readOnly)
               return true;
         }
         
         return false;
      }
      
      // since we are editable, there must be another field that can be
      // accessed
      return true;
   }
   
   /**
    * Reports if there is at least 1 focusable component that is not the
    * given component.
    *
    * @param    exclude
    *           A component to ignore or <code>null</code> if any focusable
    *           component will do.
    *
    * @return   <code>true</code> if there is at least one other component
    *           that can be focused.
    */
   @Override
   public boolean hasFocusable(Widget<O> exclude)
   {
      return hasFocusable(true, exclude) || parent().hasFocusable(exclude); 
   }
       
   /**
    * This method is called from the in-place editor when it need to initiate
    * switching to the next widget.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Container#nextFocus()
    */
   @Override
   public void nextFocus()
   {
      if (editPossible && getActiveEditor() != null)
      {
         goDown(true);
         repaint();
      }
      else
         parent().nextFocus();
   }

   /**
    * This method is called from the in-place editor when it need to initiate
    * switching to the previous widget.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Container#prevFocus()
    */
   @Override
   public void prevFocus()
   {
      if (editPossible && getActiveEditor() == null)
         switchToPrevCell();
      else
         parent().prevFocus();
   }
   
   /**
    * Returns selection mode state.
    * 
    * @return Is browse in the multiple selection mode.
    */
   public boolean isSelectionMultiple()
   {
      return (selectedRowIndices != null);
   }

   /**
    * Switch multiple selection on and off.
    *
    * @param   multiple
    *          <code>true</code> enable multiple selection,
    *          <code>false</code> - disable multiple selection.
    */
   public void setSelectionMultiple(boolean multiple)
   {
      if (browseModel == null)
         return;
      
      if (multiple && selectedRowIndices == null)
      {
         selectedRowIndices = new LinkedHashSet<>();
         if (getCurrentRow() >= 0)
         {
            selectedRowIndices.add(getCurrentRowAsSelected());
         }
         
         return;
      }
      
      if (!multiple && selectedRowIndices != null)
      {
         selectedRowIndices = null;
         return;
      }
      
      // do nothing, other cases should not have any effect
      return;
   }

   /**
    * Returns is given row selected.
    * 
    * @param   rowIndex
    *          0-based row's index in browse model.
    * 
    * @return  Is this row selected.
    */
   public boolean isRowSelected(int rowIndex)
   {
      if (selectedRowIndices == null)
      {
         return (Math.max(getCurrentRow(), 0) == rowIndex) && currRowSelected;
      }

      return selectedRowIndices.contains(new SelectedRow(rowIndex));
   }

   /**
    * Implements the FOCUSED-ROW attribute getter.
    *
    * @return 1-based index of the focused row in the viewport. If there are no rows,
    *         <code>1</code> is returned,
    */
   public int getFocusedRow()
   {
      if (getCurrentRow() < 0)
         return 1;

      return getCurrentRow() - getScrollRow() + 1;
   }

   /**
    * Returns the count of selected rows.
    * 
    * @return  The count of selected rows.
    */
   public int getSelectedRowsCount()
   {
      if (selectedRowIndices == null)
      {
         return (getCacheSize() > 0) ? 1 : 0;
      }

      return selectedRowIndices.size();
   }

   /**
    * Get the index (in the browse model) of the given selected row.
    *
    * @param   selIndex
    *          0-based position of the row in the list of selected rows. <code>-1</code> for the
    *          last selected row.
    *
    * @return  0-based index of the row in the browse model.
    *
    * @throws  com.goldencode.p2j.util.ErrorConditionException
    *          If <code>selIndex</code> is invalid or if there are no "available" rows.
    */
   public int getSelectedRowIndex(int selIndex)
   {
      return getSelectedRowIndex(selIndex, false);
   }

   /**
    * Get the index (in the browse model) of the given selected row.
    *
    * @param   selIndex
    *          0-based position of the row in the list of selected rows. <code>-1</code> for the
    *          last selected row.
    * @param   ignoreNotAvail
    *          If <code>true</code> then exception is not raised if the selected row is not
    *          available.
    *
    * @return  0-based index of the row in the browse model.
    *
    * @throws  com.goldencode.p2j.util.ErrorConditionException
    *          If <code>selIndex</code> is invalid or if there are no "available" rows and
    *          <code>ignoreNotAvail</code> parameter is not specified.
    */
   public int getSelectedRowIndex(int selIndex, boolean ignoreNotAvail)
   {
      int numSelectedRows = getSelectedRowsCount();
      if (selIndex == -1)
      {
         // DELETE-CURRENT-ROW mode
         if (numSelectedRows == 0)
            throw new ErrorConditionException(382, "");

         selIndex = numSelectedRows - 1;
      }

      if (selIndex >= numSelectedRows || selIndex < 0)
      {
         throw new ErrorConditionException(385, "");
      }

      if (!ignoreNotAvail && !isSelectedRowAvailable(selIndex))
      {
         throw new ErrorConditionException(382, "");
      }

      if (selectedRowIndices == null)
      {
         return getCurrentRow();
      }

      Iterator<SelectedRow> itRow = selectedRowIndices.iterator();
      for (int i = selIndex; i > 0; --i)
      {
         itRow.next();
      }

      return itRow.next().getRowIndex();
   }

   /**
    * Retrieve the row indexes of the selected rows in the given browse widget.
    *
    * @return  Array of 0-based row indexes of the selected rows in the browse.
    *
    * @throws  com.goldencode.p2j.util.ErrorConditionException
    *          If <code>selIndex</code> is invalid or if there are no "available" rows.
    */
   public int[] getSelectedRowIndexes()
   {
      int numSelectedRows = getSelectedRowsCount();
      if (numSelectedRows == 0)
      {
         throw new ErrorConditionException(382, "");
      }

      if (!isSelectedRowAvailable(-1))
      {
         throw new ErrorConditionException(382, "");
      }

      if (selectedRowIndices == null)
      {
         return new int[] { getCurrentRow() };
      }

      int[] res = new int[selectedRowIndices.size()];
      int i = 0;
      for (SelectedRow row : selectedRowIndices)
      {
         res[i++] = row.getRowIndex();
      }

      return res;
   }

   /**
    * Get the maximal row index (in the browse model) among the selected rows,
    *
    * @return maximal row index (in the browse model) among the selected rows or <code>-1</code>
    *         if there are no selected rows.
    */
   public int getMaximalSelectedRowIndex()
   {
      int numSelectedRows = getSelectedRowsCount();
      if (numSelectedRows == 0)
      {
         return -1;
      }

      if (selectedRowIndices == null)
      {
         return getCurrentRow();
      }

      int res = -1;
      for (SelectedRow row : selectedRowIndices)
      {
         int index = row.getRowIndex();
         if (index > res)
            res = index;
      }

      return res;
   }

   /**
    * Fires the DefaultAction event.
    */
   public void fireDefaultAction()
   {
      EventManager.postEvent(new KeyInput(this, 
                                    EventType.KEY_PRESSED, Keyboard.SE_DEFAULT_ACTION));
   }

   /**
    * Fires the ValueChanged event.
    */
   public void fireValueChanged()
   {
      ConfigHelper.setModified(config, true);

      ThinClient.setValueChangedEvent(new KeyInput(this,
                                                   EventType.KEY_PRESSED,
                                                   Keyboard.SE_VALUE_CHANGED,
                                                   true));
   }
   
   /**
    * Return status of the updates of the current row.
    * 
    * @return  <code>true</code> if current row is modified and
    *          <code>false</code> otherwise.
    */
   public boolean isCurrentRowModified()
   {
      return (currRowUpdates != null);
   }

   /**
    * Set number of rows to display data.
    * 
    * @param   down
    *          Number of data rows.
    */
   public void setDown(int down)
   {
      if (browseModel == null)
         return;
      
      // account for the title
      if (config.box)
         down += 2;

      // account for labels
      if (drawLabels)
         down += ilLabelsHeight + (config.noUnderline ? 0 : 1);

      setSize(width(), down);
   }

   /**
    * Return <code>true</code> if row which has focus is selected.
    * 
    * @return  <code>true</code> if row which has focus is selected and
    *          <code>false</code> otherwise.
    */
   public boolean isFocusedRowSelected()
   {
      return isRowSelected(getCurrentRow());
   }
   
   /**
    * Return number of potential data rows in the <code>Browse</code>.
    * Amount of available data can be less that number of rows which
    * potentially can be displayed by the <code>Browse</code>.
    * 
    * @return  Number of potential rows.
    */
   public int getDown()
   {
      return (int) ilPageRowsCount;
   }

   /**
    * Set current column to specified one.
    * 
    * @param   currentColumn
    *          Reference to <code>BrowseColumn</code> instance of specified
    *          column.
    */
   public void setCurrentColumn(BrowseColumnConfig currentColumn)
   {
      BrowseColumnConfig[] cols = browseModel.getColumns();
      
      for (int i = 0; i < cols.length; i++)
      {
         if (currentColumn == cols[i])
         {
            FillIn<O,?> editor = getActiveEditor();
            
            if (editor != null)
               stopEdit();
            
            currColumn = i;
            
            if (editor != null)
               startEdit();
         }
      }
   }

   /**
    * Get currently used <code>BrowseModel</code>.
    * 
    * @return  Reference to <code>BrowseModel</code> used by
    *        <code>Browse</code>.
    */
   public BrowseModel getModel()
   {
      return browseModel;
   }

   /**
    * Return reference to previous column relative to given column.
    * 
    * @param   column
    *          Base column.
    * 
    * @return  Reference to next column relative to given.
    */
   public BrowseColumnConfig getPrevColumn(BrowseColumnConfig column)
   {
      BrowseColumnConfig[] cols = browseModel.getColumns();

      for (int i = 0; i < cols.length; i++)
      {
         // found it
         if (column == cols[i])
         {
            if ((i + 1) < cols.length)
               return cols[i + 1];
            return null;
         }
      }

      return null;
   }

   /**
    * Return reference to next column relative to given column.
    * 
    * @param   column
    *          Base column.
    * 
    * @return  Reference to next column relative to given.
    */
   public BrowseColumnConfig getNextColumn(BrowseColumnConfig column)
   {
      BrowseColumnConfig[] cols = browseModel.getColumns();

      for (int i = 0; i < cols.length; i++)
      {
         // found it
         if (column == cols[i])
         {
            if ((i - 1) > 0)
               return cols[i - 1];
            return null;
         }
      }

      return null;
   }

   /**
    * Get widget IDs of the browse column widgets.
    *
    * @return array of widget IDs of the browse column widgets.
    */
   public int[] getIdsOfColumnWidgets()
   {
      if (browseModel == null)
         return new int[0];

      BrowseColumnConfig[] columns = browseModel.getColumns();
      int[] res = new int[columns.length];

      for (int i = 0; i < columns.length; i++)
         res[i] = columns[i].id.asInt();

      return res;
   }

   /**
    * Determines if the specific selected row was available (could be fetched) at the moment of
    * its selection.
    *
    * @param  selIndex
    *         0-based position of the row in the list of selected rows. If <code>-1</code> is
    *         passed then all of the selected rows are checked for availability.
    *
    * @return <code>true</code> if the specific selected row was available (could be fetched) at
    *         the moment of its selection.
    *         If <code>-1</code> is passed then <code>true</code> is returned if at least one
    *         of the selected rows is available.
    */
   public boolean isSelectedRowAvailable(int selIndex)
   {
      if (selectedRowIndices == null)
      {
         if (selIndex > 0)
            return false;

         return currentRowAvail;
      }
      else
      {
         if (selIndex >= selectedRowIndices.size())
            return false;

         boolean checkAllRows = selIndex < 0;

         Iterator<SelectedRow> iter = selectedRowIndices.iterator();
         int i = 0;
         while (iter.hasNext())
         {
            SelectedRow row = iter.next();
            if (checkAllRows)
            {
               if (row.isRowAvailable())
                  return true;
            }
            else
            {
               if (selIndex == i)
                  return row.isRowAvailable();
               i++;
            }
         }

         // no available rows
         return false;
      }
   }

   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(BrowseConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
      
      if (config.heightChars == 0 && config.widthChars == 0)
      {
         setDown(config.down * (int) config.rowHeightChars);
      }

      setSelectionMultiple(config.multiple);
      setReadOnly(config.readOnly);

      if (config.queryId != -1)
      {
         setQuery(config.queryId);
      }
   }

   /**
    * Return <code>BrowseColumn</code> for specified 0-based index.
    * 
    * @param   columnId
    *          Column index.
    * 
    * @return  <code>BrowseColumn</code> instance or <code>null</code> in
    *          case of invalid index.
    */
   BrowseColumnConfig getColumn(int columnId)
   {
      BrowseColumnConfig[] cols = browseModel.getColumns();
      
      if (columnId >= 0 && columnId < cols.length)
         return cols[columnId];
      
      return null;
   }

   /**
    * Returns the active in-place editor if edit is in progress. 
    *
    * @return Reference to editor or <code>null</code>.
    */
   @SuppressWarnings("unchecked")
   protected FillIn<O,?> getActiveEditor()
   {
      return (FillIn<O,?>) get(0);
   }

   /**
    * Trigger row display events, if any row requires it.
    */
   @SuppressWarnings("unchecked")
   protected void triggerRowDisplayEvents()
   {
      // avoid recursion
      if (rowDisplayTriggered)
         return;

      ThinClient client = ThinClient.getInstance();
      BrowseColumnConfig[] columns = browseModel.getColumns();
      rowDisplayTriggered = true;
      int oldRow = getCurrentRow();

      // execute any ROW-DISPLAY triggers
      try
      {
         //int offset = getScrollRow() - getDcStartRow();
         //int maxVal = (int) Math.min(offset + ilPageRowsCount, rowDisplayed.length);
         int size = getCacheSize();
         for (int i = 0; i <  size; i++)
         {
            if (rowDisplayed[i])
               continue;

            rowDisplayed[i] = true;

            int rowIdx = getCachedRowNumber(i);

            browseModel.setCurrentRow(rowIdx);

            // restore the colors to their default values, so the trigger
            // may overwrite them
            for (int k = 0; k < columns.length; k++)
            {
               WidgetId cid = columns[k].id;
               BrowseColumn<O> bcol = (BrowseColumn<O>) client.getWidget(cid);
               Color color = colors[i] == null ? null : colors[i][k];
               bcol.config().columnDColor = color == null ? Color.DEFAULT : color;
            }

            client.invokeRowDisplayTrigger(this.config().id.asInt());
            
            // set the colors, as they might have been modified by the 
            // trigger
            for (int k = 0; k < columns.length; k++)
            {
               WidgetId cid = columns[k].id;
               BrowseColumn<O> bcol = (BrowseColumn<O>) client.getWidget(cid);
               if (colors[i] == null)
                  colors[i] = new Color[columns.length];

               colors[i][k] = bcol.config().columnDColor;
            }
         }
      }
      finally
      {
         // ensure cleanup is always possible
         browseModel.setCurrentRow(oldRow);
      }
   }

   /**
    * Setups scrolling to fully (as possible) show the given column.
    * 
    * @param   cellColumn
    *          The cell's column index in model.
    */
   protected void columnToView(int cellColumn)
   {
      int newScrollColumn = scrollColumn;
      
      if (cellColumn == 0)
      {
         newScrollColumn = 0;
      }
      else if (cellColumn != newScrollColumn)
      {
         BrowseColumnConfig[] cols = browseModel.getColumns();
         
         if (cellColumn > newScrollColumn)
         {
            newScrollColumn = cellColumn;
         }
         else
         {
            double usedX = cols[newScrollColumn].widthChars;
            for (int i = cellColumn; i < newScrollColumn; ++i)
            {
               usedX += cols[i].widthChars + 1;
            }

            int availableWidth = getAvailableWidth();
            // safe, comparison of scaled doubles
            while ((usedX = Coordinate.scale(usedX)) > availableWidth)
            {
               usedX -= cols[newScrollColumn].widthChars + 1;
               --newScrollColumn;
               if (newScrollColumn == cellColumn)
               {
                  // new scroll column may be wider than all browse
                  break;
               }
            }
         }

         if (newScrollColumn < ilFirstScrollCol)
         {
            newScrollColumn = 0;
         }
      }
      
      if (newScrollColumn != scrollColumn)
      {
         scrollColumn = newScrollColumn;
         rcLabels = null;
      }
   }

   /**
    * Creates an editor for the current cell and adds it to the browse. Adjusts selection data if
    * necessary.
    */
   @SuppressWarnings("unchecked")
   protected void startEdit()
   {
      startEdit(true);
   }

   /**
    * Finishes edit and destroys an editor.
    *
    * @return  <code>true</code> if editing was stopped;  <code>false</code>
    *          if we were not in edit mode.
    */
   protected boolean stopEdit()
   {
      FillIn<O,?> editor = getActiveEditor();

      if (editor == null || !editor.update())
         return false;

      BaseDataType var = editor.getValue();
      updateColumn(currColumn, var);

      remove(editor);
      screen().getRegistry().removeWidget(editor.config().id);

      Window.setInsertMode(Window.INSERT_MODE_OVR);

      return true;
   }

   /**
    * Creates an editor for the current cell and adds it to the browse. Adjusts selection data if
    * necessary.
    *
    * @param updateSelection
    *        For multi-selection browse: if <code>true</code>, current row becomes the only
    *        selected row.
    */
   @SuppressWarnings("unchecked")
   private void startEdit(boolean updateSelection)
   {
      if (getActiveEditor() != null)
         return;

      BrowseColumnConfig[] cols = browseModel.getColumns();

      if (cols[currColumn].readOnly)
         return;

      FillInConfig cfg = new FillInConfig(WidgetId.nextID());

      if (cols[currColumn].format != null)
         cfg.format = cols[currColumn].format; // direct access

      cfg.dataType   = cols[currColumn].dataType; // direct access
      cfg.pfcolor    = config.pfcolor;
      cfg.dcolor     = config.dcolor;
      cfg.autoReturn = cols[currColumn].autoReturn;
      cfg.autoZap    = cols[currColumn].autoZap;
      cfg.help       = cols[currColumn].help;

      // safe, comparison of scaled doubles
      double width = Math.min(cols[currColumn].widthChars, ilPageWidth);
      cfg.widthChars = width;
      
      double x = 2;
      double y = getViewportRow();
      
      if (config.box)
      {
         ++x;
         ++y;
      }

      if (drawLabels)
      {
         y += ilLabelsHeight + (config.noUnderline ? 0 : 1);
      }
      
      if (scrollColumn == 0)
      {
         for (int i = 0; i < currColumn; ++i)
         {
            x += cols[i].widthChars + 1;
         }
      }
      else
      {
         x += ilPageWidth - width;
         for (int i = scrollColumn; i > currColumn; --i)
         {
            x -= cols[i].widthChars + 1;
         }
      }

      cfg.clientColumn = x; // direct access
      cfg.clientRow = y; // direct access
      cfg.visible = false; // direct access
      
      WidgetFactory<O> factory = screen().getFactory();
      WidgetId editorId = WidgetId.nextID();
      FillIn<O,FillInConfig> editor = factory.create(FillIn.class, editorId, cfg);
      screen().getRegistry().addWidget(editorId, (Widget) editor);
      editor.initialize(editorId, cfg);
      
      // reset underline attribute and force Browse colors
      editor.setBrowse(this);
      editor.setValue(getValue(currColumn));
      editor.setUnderline(false);
      add(editor);

      editor.setEnabled(true);
      editor.setVisible(true);
      editor.activate();
      
      // the following call is an internal focus change and is safe
      setFocus(editor);
      
      EventManager.postEvent(new FocusEvent(editor, EventType.FOCUS_GAINED));
      columnHelp  = cols[currColumn].help;

      if (columnHelp != null)
      {
         ThinClient.setHelp(columnHelp);
      }

      if (selectedRowIndices != null && updateSelection)
      {
         selectedRowIndices.clear();
         if (getCurrentRow() >= 0)
         {
            selectedRowIndices.add(getCurrentRowAsSelected());
         }
      }
   }
   
   /**
    * Update specified cell in current row.
    * 
    * @param   index
    *          0-based index of the cell to update.
    * @param   var
    *          New value of the cell.
    * 
    * @return  <code>true</code> if column value was changed.
    */
   private boolean updateColumn(int index, BaseDataType var)
   {
      if (getCacheSize() == 0)
         return false;

      boolean res = true;

      if (dcGetRow(getCurrentRow())[index].equals(var))
      { 
         // no real edit
         
         res = false;
         
         if (currRowUpdates != null)
            currRowUpdates[index] = null;
      }
      else
      {
         if (currRowUpdates == null)
            currRowUpdates = new BaseDataType[browseModel.getColumns().length];

         currRowUpdates[index] = var;
      }
      
      return res;
   }
   
   /**
    * Switches in-line editor to the previous editable cell. Initiates commit
    * of the current row if necessary.
    */
   void switchToPrevCell()
   {
      BrowseColumnConfig[] cols = browseModel.getColumns();

      int prevColumn;
      for (prevColumn = currColumn - 1; prevColumn >= 0; 
           --prevColumn)
      {
         if (!cols[prevColumn].readOnly)
            break;
      }

      do
      {
         if (prevColumn >= 0)
         {
            stopEdit();
            currColumn = prevColumn;
            startEdit();
            break;
         }

         if (!leaveRow())
            break;

         int oldCurrRow = getCurrentRow();
         goUp(false);

         if (oldCurrRow != getCurrentRow())
         {
            currColumn = cols.length;

            while (cols[--currColumn].readOnly)
            {
               // empty loop body
            }
         }

         startEdit();
         fireValueChanged();
      }
      while (false);
   
      repaint();
   }
   
   /**
    * Switches in-line editor to the next editable cell. Initiates commit of
    * the current row if necessary.
    */
   void switchToNextCell()
   {
      BrowseColumnConfig[] cols = browseModel.getColumns();

      int nextColumn;
      for (nextColumn = currColumn + 1; nextColumn < cols.length; ++nextColumn)
      {
         if (!cols[nextColumn].readOnly)
            break;
      }

      do
      {
         if (nextColumn < cols.length)
         {
            stopEdit();
            currColumn = nextColumn;
            startEdit();
            break;
         }

         if (!leaveRow())
            break;
   
         int oldCurrRow = getCurrentRow();
         goDown(false);

         if (oldCurrRow != getCurrentRow())
         {
            currColumn = 0;

            while (cols[currColumn].readOnly)
               ++currColumn;
         }
   
         startEdit();
         fireValueChanged();
         
      } while (false);

      repaint();
   }

   /**
    * Saves the current row into cache and data model.
    *
    * @param  updateModel
    *         <code>true</code> if data model is updated. Otherwise only cache is updated.
    *
    * @return Is data checking and saving was successful.
    */
   private boolean saveCurrentRow(boolean updateModel)
   {
      if (currRowUpdates == null)
      {
         return true;
      }

      int len = 0;
      for (int i = 0; i < currRowUpdates.length; ++i)
      {
         if (currRowUpdates[i] != null)
         {
            ++len;
         }
      }

      if (len == 0)
      {
         return true;
      }

      int[] indices = new int[len];
      BaseDataType[] values = new BaseDataType[len];
      
      int p = 0;
      for (int i = 0; i < currRowUpdates.length; ++i)
      {
         BaseDataType update = currRowUpdates[i];
         if (update != null)
         {
            indices[p] = i;
            values[p] = update;
            ++p;
         }
      }

      try
      {
         dcUpdateRow(getCurrentRow(), indices, values, updateModel);
      }
      catch (DataValidationException ex)
      {
         // Enable internal, row-level undo.
         awaitingEndError = true;
         StringBuilder buf = new StringBuilder("Press ");
         buf.append(Keyboard.kbLabel(Keyboard.END_ERROR));
         buf.append(" to undo changes to the current row in BROWSE. (5133)");
         // TODO:  embed browse name in message text after 'BROWSE';
         // currently, config.getName() returns null.
         //buf.append(" to undo changes to the current row in BROWSE ");
         //buf.append(config.getName());
         //buf.append(". (5133)");
         window().message(buf.toString());
         return false;
      }

      currRowUpdates = null;
      return true;
   }

   /**
    * Finishes editing and commits row to model if some updates are available. Restores editor if
    * commit has failed.
    *
    * @return  <code>true</code> if all actions were successful.
    */
   private boolean releaseCurrentRow()
   {
      return releaseCurrentRow(true);
   }

   /**
    * Finishes editing and commits row to model if some updates are available. Restores editor if
    * commit has failed.
    *
    * @param   updateModel
    *          <code>true</code> if data model is updated. Otherwise only cache is updated.
    *
    * @return  <code>true</code> if all actions were successful.
    */
   private boolean releaseCurrentRow(boolean updateModel)
   {
      FillIn<O,?> editor = getActiveEditor();
      
      if (editor == null)
         return true;

      boolean status = stopEdit();
      
      if (!status)
         return false;
      
      return saveCurrentRow(updateModel);
   }

   /**
    * Returns the data row with the given index. The row must already be in
    * the cache. 
    *
    * @param  rowIndex
    *         Row's index in model.
    *
    * @return The data row from the cache.
    */
   protected BaseDataType[] dcGetRow(int rowIndex)
   {
      BaseDataType[] numberedRow = dcData[dcGetRowIndex(rowIndex)];
      BaseDataType[] row = new BaseDataType[numberedRow.length - 1];
      System.arraycopy(numberedRow, 1, row, 0, row.length);
      return row;
   }

   /**
    * Updates row in the cache and the model. Row must already be in cache.
    *
    * @param  rowIndex
    *         0-based row's index in model.
    * @param  cellIndices
    *         Indexes of cells to update.
    * @param  cellValues
    *         New values of cells to update.
    * @param  updateModel
    *         <code>true</code> if data model is updated. Otherwise only cache is updated.
    *
    * @throws DataValidationException
    *         New values don't passed database validation rules.
    */
   private void dcUpdateRow(int            rowIndex,
                            int[]          cellIndices,
                            BaseDataType[] cellValues,
                            boolean        updateModel)
   throws DataValidationException
   {
      BaseDataType[] row = dcData[dcGetRowIndex(rowIndex)];

      updateModel = updateModel && !config.noAssign;
      if (updateModel)
         browseModel.updateRow(rowIndex, cellIndices, cellValues);

      for (int i = 0; i < cellIndices.length; ++i)
      {
         row[cellIndices[i] + 1] = cellValues[i];
      }
   }

   /**
    * Process key event.
    * 
    * @param   event
    *          Reference to keyboard event.
    */
   @Override
   public void onKeyPressed(KeyInput event)
   {
      int key = event.actionCode();
      
      FillIn<O,?> editor = getActiveEditor();
      
      if (key == Keyboard.KA_TAB)
      {
         event.consume();

         if (!invokeRowLeaveTrigger())
            return;

         if (releaseCurrentRow())
         {
            if (editor != null)
               startEdit(false); // editor must be preserved in edit mode

            // the following request should be conditional
            ThinClient.nextFocus(Browse.this);
         }
         return;
      }
      
      if (key == Keyboard.KA_BACK_TAB && editor == null)
      {
         event.consume();

         if (!invokeRowLeaveTrigger())
            return;

         if (releaseCurrentRow())
         {
            //TODO: editor can only be null at this point (see 'if' 5 lines
            // above. Requires deeper investigation.
            //if (editor != null)
            //   startEdit(); // editor must be present in edit mode
            
            // the following request should be conditional
            ThinClient.previousFocus(Browse.this);
         }
         return;
      }
      
      if (getCacheSize() == 0)
      { 
         // all other keys requires the presence of current row
         return;
      }
      
      boolean invalid = true;
      boolean editing = editor != null;
      
      do
      {
         if (key == Keyboard.KA_CURSOR_LEFT)
         {
            if (scrollColumn == 0)
            {
               invalid = false;
               break;
            }
               
            --scrollColumn;
            if (scrollColumn < ilFirstScrollCol)
            {
               scrollColumn = 0;
            }

            rcLabels = null;
            break;
         }
            
         if (key == Keyboard.KA_CURSOR_RIGHT)
         {
            int colsCount = browseModel.getColumns().length - 1;
            if (scrollColumn >= colsCount)
            {
               invalid = false;
               break;
            }
               
            if (scrollColumn == 0)
            {
               if (ilFirstScrollCol > colsCount)
               {
                  invalid = false;
                  break;
               }
               scrollColumn = ilFirstScrollCol;
            }
            else
            {
               ++scrollColumn;
            }
               
            rcLabels = null;
            break;
         }
            
         if (key == Keyboard.KA_CURSOR_UP)
         {
            goUp(true);
            break;
         }
            
         if (key == Keyboard.KA_CURSOR_DOWN)
         {
            goDown(true);
            break;
         }
            
         if (key == Keyboard.KA_HOME ||
             key == Keyboard.KA_LEFT_END)
         {
            if (editing && !leaveRow())
               break;

            int oldCurrRow = getCurrentRow();
            boolean anyRows = scrollToRow(0, false, true);
            if (anyRows)
               setCurrentRow(getCachedRowNumber(0));

            if (editing)
            {
               startEdit();
            }

            if (!isSelectionMultiple() && oldCurrRow != getCurrentRow())
               fireValueChanged();
               
            break;
         }
            
         if (key == Keyboard.KA_END ||
             key == Keyboard.KA_RIGHT_END)
         {
            if (editing && !leaveRow())
               break;

            int oldCurrRow = getCurrentRow();
            int bigNumber = Integer.MAX_VALUE - 1000; // decreased to avoid overflow
            boolean anyRows = scrollToRow(bigNumber, true, true);
            if (anyRows)
               incrementalScroll( -(ilPageRowsCount - 1));
            resetCacheStructures();

            if (editing)
            {
               startEdit();
            }

            if (!isSelectionMultiple() && oldCurrRow != getCurrentRow())
               fireValueChanged();

            break;
         }

         if (key == Keyboard.KA_PAGE_UP)
         {
            // In 4GL if the previous page is fully deleted, you have to press Page Up twice in
            // order to move to the page before the deleted one. But eventually gaps of deleted
            // rows can be closed and you'll have to press Page Up only once.
            // In P2J for now you'll have to press Page Up only once for the beginning.
            // TODO can be implemented after gap closing mechanism is implemented

            if (editing && !leaveRow())
               break;

            final int oldScrollRow = getScrollRow();
            int oldCurrRow = getCurrentRow();
            int newScrollRow;
            if (oldScrollRow > 0)
            {
               int virtualRow = Math.max(oldScrollRow - ilPageRowsCount, 0);
               int viewportRow = getViewportRow();

               while (true)
               {
                  boolean anyRows = scrollToRow(virtualRow, false, true);
                  newScrollRow = getScrollRow();

                  if (!anyRows            ||
                      virtualRow == 0     ||
                      newScrollRow < oldScrollRow)
                  {
                     // The result set is empty OR
                     // We reached the first row OR
                     // We found a page with has an least one row.
                     break;
                  }

                  virtualRow = Math.max(virtualRow - ilPageRowsCount, 0);
               }

               viewportRow = newScrollRow == oldScrollRow ? 0 :
                             Math.min(getCacheSize() - 1, viewportRow);
               setCurrentRow(getCachedRowNumber(viewportRow));
            }
            else
            {
               setCurrentRow(0);
            }

            if (editing)
            {
               startEdit();
            }

            if (!isSelectionMultiple() && oldCurrRow != getCurrentRow())
               fireValueChanged();

            break;
         }

         if (key == Keyboard.KA_PAGE_DOWN)
         {
            if (editing && !leaveRow())
               break;

            int oldCurrRow = getCurrentRow();
            int viewportRow = getViewportRow();

            if (getCacheSize() < ilPageRowsCount ||
                !incrementalScroll(ilPageRowsCount))
            {
               // no more rows after
               int newRowModelIndex = getCachedRowNumber(getCacheSize() - 1);
               setCurrentRow(newRowModelIndex);
            }
            else
            {
               int row = Math.min(viewportRow, getCacheSize() - 1);
               setCurrentRow(getCachedRowNumber(row));
            }

            if (editing)
            {
               startEdit();
            }

            if (!isSelectionMultiple() && oldCurrRow != getCurrentRow())
               fireValueChanged();

            break;
         }

         if (key == Keyboard.KA_REPLACE)
         {
            if (editing)
            {
               if (!invokeRowLeaveTrigger())
                  break;

               setReadOnly(true);
               releaseCurrentRow();
               invalid = true;
            }
            else
            {
               if (!editPossible)
               {
                  invalid = false;
                  break;
               }

               setReadOnly(false);
            }

            break;
         }

         if (key == Keyboard.SE_ENTRY)
         {
            fireDefaultAction();
            break;
         }

         if (editing && (key == Keyboard.KA_BACK_TAB))
         {
            switchToPrevCell();
            break;
         }

         if (editing && (key == Keyboard.KA_EDITOR_TAB))
         {
            switchToNextCell();
            break;
         }

         if (editing && key == Keyboard.KA_END_ERROR && awaitingEndError)
         {
            awaitingEndError = false;
            undoValue();
            break;
         }

         return; // ignoring other keys
      }
      while (false);

      event.consume(); // consume all handled keys
      
      if (invalid)
      {
         draw();
         requestSync();
      }
      else if (getActiveEditor() != null)
      {
         editor = getActiveEditor(); // editor may be recreated
         editor.draw();
         requestSync();
      }
   }
   
   /**
    * Intercept special events. Using regular keyTyped/keyPressed may not work
    * if component is not in focus.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Widget#processEvent(com.goldencode.p2j.ui.client.event.Event)
    */
   @Override
   public void processEvent(Event evt) 
   {
      if (evt instanceof KeyInput)
      {
         KeyInput ki = (KeyInput) evt;
         int action = ki.actionCode();
         
         if(action == Keyboard.SE_VALUE_CHANGED ||
            action == Keyboard.SE_ENTRY)
         {
            // if widget is not initialized then do nothing
            if (getActiveEditor() != null || browseModel == null)
            {
               return;
            }

            // ignore VALUE-CHANGED events generated by Browse, 
            // they are for external consumers, all internal actions
            // are already performed 
            if (action == Keyboard.SE_VALUE_CHANGED && ki.isSpecial())
            {
               return;
            }
            
            refreshScrollRow(false, false);
            return;
         }
         
         onKeyPressed(ki);
         boolean consumed = ki.isConsumed();
         
         if (consumed                     || 
             action == Keyboard.KA_RETURN || 
             action == Keyboard.keyCode(" "))
         {
            // if the key was processed by the browse, clear the message area
            // if space or return was pressed, clear the message area
            window().clearPutScreenInMessageArea(-1);
         }
         
         if (consumed)
            return;
      }
      
      super.processEvent(evt);
   }
   
   /**
    * Handles ' ' key in R/O mode: fires ValueChanged event. Additionally
    * inverts selection status of the current line in multiple selection
    * mode.
    *
    * @see com.goldencode.p2j.ui.client.event.listener.KeyListener#onKeyTyped(KeyInput)
    */
   @Override
   public void onKeyTyped(KeyInput event)
   {
      if (getCacheSize() == 0)
      {
         return;
      }
      
      if (event.keyCode() == ' ' && getActiveEditor() == null)
      {
         if (selectedRowIndices != null)
         {
            SelectedRow currRow = getCurrentRowAsSelected();
            
            if (selectedRowIndices.contains(currRow))
            {
               selectedRowIndices.remove(currRow);
            }
            else
            {
               selectedRowIndices.add(currRow);
            }
         }
         
         fireValueChanged();
         draw();
         requestSync();
      }
   }

   /**
    * Set the current row.
    *
    * @param   currRow
    *          New current row (0-based index in the browse model).
    */
   protected void setCurrentRow(int currRow)
   {
      setCurrentRow(currRow, true);
   }

   /**
    * Set the current row.
    *
    * @param   currRow
    *          New current row (0-based index in the browse model).
    * @param   allowDeletedRow
    *          If <code>true</code>, then <code>currRow</code> becomes the current row even
    *          if it doesn't exist in the result set anymore. If <code>false</code> then the next
    *          available row becomes the current row if the <code>currRow</code> doesn't exist.
    */
   private void setCurrentRow(int currRow, boolean allowDeletedRow)
   {
      setCurrentRow(currRow, true, allowDeletedRow);
   }

   /**
    * Set the current row.
    * 
    * @param   currRow
    *          New current row (0-based index in the browse model).
    * @param   updateSelection
    *          <code>true</code> if selection should be adjusted.
    * @param   allowDeletedRow
    *          If <code>true</code>, then <code>currRow</code> becomes the current row even
    *          if it doesn't exist in the result set anymore. If <code>false</code> then the next
    *          available row becomes the current row if the <code>currRow</code> doesn't exist.
    */
   protected void setCurrentRow(int currRow, boolean updateSelection, boolean allowDeletedRow)
   {
      if (browseModel == null || getCacheSize() == 0)
      {
         this.currentRow = 0;
         currentRowAvail = false;
      }
      else
      {
         int availRow = browseModel.setCurrentRow(currRow);
         this.currentRow = allowDeletedRow ? currRow : availRow;
         currentRowAvail = availRow == currRow;

         if (updateSelection)
            currRowSelected = true;
      }
   }
   
   /**
    * Implements the <b>SET-REPOSITIONED-ROW() </b> <b>BROWSE </b> widget
    * method.
    * 
    * @param   n
    *          0-based row number where the new record is displayed
    * @param   conditional
    *          if <code>true</code>, tries to reposition to the browse
    *          viewport existing row first
    */
   public void selectRepositionedRow(int n, boolean conditional)
   {
      // TODO: implement it
   }

   /**
    * Get text value of the given column in current row.
    * 
    * @param   ordinal
    *          The column ordinal.
    * 
    * @return  The text value of the given column in current row.
    */
   public String getColumnText(int ordinal)
   {
      if (ordinal == currColumn && getActiveEditor() != null)
         return getActiveEditor().getText(); 
         
      columnToView(ordinal);
      
      String fmt = browseModel.getColumns()[ordinal].format;
      BaseDataType value = null;
      
      value = getValue(ordinal);

      if (value == null)
         return "";

      return (fmt == null) ? 
                            character.valueOf(value).getValue():
                            character.valueOf(value, fmt).getValue(); 
   }

   /**
    * Notify the browse that the row was deleted from the result set.
    *
    * @param rowIndex
    *        0-based index of the deleted row.
    */
   public void notifyRowDeleted(int rowIndex)
   {
      if (selectedRowIndices != null)
      {
         if (rowIndex < 0)
         {
            selectedRowIndices.clear();
         }
         else
         {
            SelectedRow[] res = selectedRowIndices.toArray(
                                 new SelectedRow[selectedRowIndices.size()]);
            selectedRowIndices.clear();

            for (SelectedRow selectedRow : res)
            {
               if (selectedRow.getRowIndex() < rowIndex)
                  selectedRowIndices.add(selectedRow);
               else if (selectedRow.getRowIndex() > rowIndex)
                  selectedRowIndices.add(new SelectedRow(selectedRow.getRowIndex() - 1,
                                                         selectedRow.isRowAvailable()));
            }
         }

         repaint();
      }
   }

   /**
    * Get value for the specified column.
    * 
    * @param   index
    *          The 0-based column index.
    *          
    * @return  Column value.
    */
   private BaseDataType getValue(int index)
   {
      if (currRowUpdates != null && currRowUpdates[index] != null)
         return currRowUpdates[index];
      
      if (getCacheSize() == 0)
         return null;

      return dcGetRow(getCurrentRow())[index];
   }

   /**
    * Get value of the given column in current row.
    * 
    * @param   ordinal
    *          The column ordinal.
    * 
    * @return  The value of the given column in current row.
    */
   public BaseDataType getColumnValue(int ordinal)
   {
      if (ordinal == currColumn && getActiveEditor() != null)
         return getActiveEditor().getValue(); 
         
      BaseDataType value = getValue(ordinal);

      if (value == null)
         return new character();

      return value; 
   }

   /**
    * Set value for the given column in current row.
    * 
    * @param   ordinal
    *          The column ordinal.
    * @param   value         
    *          The value for the cell.
    */
   public void setColumnValue(int ordinal, BaseDataType value)
   {
      if (config.readOnly ||
          browseModel.getColumns()[ordinal].readOnly)
      {
         // update cache without marking it as changes to commit
         BaseDataType[] row = dcGetRow(getCurrentRow());
         row[ordinal].assign(value);
         repaint();
         return;
      }

      if (ordinal == currColumn && getActiveEditor() != null)
      {
         getActiveEditor().setValue(value);
         return;
      }

      // update currRowUpdates
      updateColumn(ordinal, value);
      repaint();
   }

   /**
    * Check if the value in specified column in current row is changed. 
    * 
    * @param   ordinal
    *          The column ordinal.
    *          
    * @return  <code>true</code> if value is changed.
    */
   public boolean isModified(int ordinal)
   {
      if (ordinal == currColumn && getActiveEditor() != null)
         return getActiveEditor().isModified(); 
         
      columnToView(ordinal);

      if (currRowUpdates != null && currRowUpdates[ordinal] != null)
         return true;

      return false;
   }

   /**
    * Refresh the browse cache and redisplay the browse.
    *
    * @param deleteCurrentRowMode
    *        <code>true</code> if this function is used for update on DELETE-CURRENT-ROW.
    *        <code>false</code> if this function is used for update on REFRESH.
    */
   public void refresh(boolean deleteCurrentRowMode)
   {
      boolean editing = getActiveEditor() != null;
      if (editing)
      {
         undoValue();
         stopEdit();
      }

      int currRow = getCurrentRow();
      int viewportRow = getViewportRow();

      scrollToRow(getCachedRowNumber(0), false, false);

      int cacheSize = getCacheSize();
      if (cacheSize > 0)
      {
         boolean found = deleteCurrentRowMode || isRowInView(currRow);

         boolean fireValueChanged = false;
         if (found)
         {
            // Old current row is in the viewport - preserve viewport row number.
            int newViewportRow = Math.min(viewportRow, getCacheSize() - 1);
            setCurrentRow(getCachedRowNumber(newViewportRow));

            if (viewportRow != newViewportRow)
               fireValueChanged = true;
         }
         else
         {
            // Old current row is not in the viewport - use the scroll row as the current row.
            scrollToRow(getCachedRowNumber(0), true);
            fireValueChanged = true;
         }

         if (editing)
            startEdit();

         if (!isSelectionMultiple() && fireValueChanged)
            fireValueChanged();
      }
   }

   /**
    * Show widget.
    */
   @Override
   public void show()
   {
      if (!isVisible())
         refresh(false);
      super.show();
   }
   
   /**
    * Execute pending refresh when REFRESHABLE is turned on. Includes pending query repositions
    * and reopenings.
    *
    * @param   queryReopened
    *          <code>true</code> if there is a pending query reopening.
    * @param   repositionBeforeReopen
    *          <code>true</code> if the last pending reposition was <i>before</i> the last pending
    *          reopening.
    * @param   repositionedRow
    *          0-based index of the last row the query was repositioned to. <code>-1</code> if
    *          there is no repositioned row.
    */
   public void executePendingRefresh(boolean queryReopened,
                                     boolean repositionBeforeReopen,
                                     int repositionedRow)
   {
      if (browseModel != null)
      {
         boolean editing = getActiveEditor() != null;
         if (queryReopened)
         {
            if (selectedRowIndices != null)
               selectedRowIndices.clear();

            if (editing)
            {
               undoValue();
               stopEdit();
            }

            if (repositionedRow >= 0 && !repositionBeforeReopen)
            {
               // This initializes rowCountAssessment for proper future scrolling.
               scrollToRow(0, false, false);
               if (editing)
                  stopEdit();

               scrollToRow(repositionedRow, true);
            }
            else
            {
               int viewportRow = getViewportRow();
               scrollToRow(0);
               if (editing)
                  stopEdit();

               if (repositionedRow >= 0 && isRowInView(repositionedRow))
               {
                  setCurrentRow(repositionedRow);
               }
               else
               {
                  viewportRow = Math.min(getCacheSize() - 1, viewportRow);
                  setCurrentRow(getCachedRowNumber(viewportRow));
               }
            }

            if (editing && getCacheSize() > 0)
               startEdit();
         }
         else if (repositionedRow >= 0)
         {
            if (!editing)
            {
               if (!isSelectionMultiple())
               {
                  // read-only single selection browse
                  scrollToRow(getCurrentRow(), false, false);
                  if (isRowInView(repositionedRow))
                     setCurrentRow(repositionedRow);
                  else
                     setCurrentRow(getCachedRowNumber(0));
               }
               else
               {
                  // read-only multi selection browse
                  int lastSelectedRow = getLastAvailableSelectedRow();
                  if (lastSelectedRow == -1)
                  {
                     scrollToRow(repositionedRow, true, false);
                  }
                  else
                  {
                     int viewportRow = getViewportRow();
                     scrollToRow(lastSelectedRow, false, false);
                     if (isRowInView(repositionedRow))
                     {
                        setCurrentRow(repositionedRow);
                     }
                     else
                     {
                        viewportRow = Math.min(getCacheSize() - 1, viewportRow);
                        setCurrentRow(getCachedRowNumber(viewportRow));
                     }
                  }
               }
            }
            else
            {
               // editable browse
               undoValue();
               stopEdit();
               scrollToRow(repositionedRow, true, false);
               if (getCacheSize() > 0)
                  startEdit();
            }
         }
      }
   }
   
   /**
    * Refresh scroll row in the browse and redisplay. Unless specific parameters are passed, the
    * current row becomes the scroll row.
    *
    * @param leaveOnRowInView
    *        If the current row is in the viewport, do not update the scroll row.
    * @param preserveViewportRow
    *        Try to preserve the current viewport row (i.e. the row at the same position in
    *        the viewport becomes the current row after refresh).
    */
   public void refreshScrollRow(boolean leaveOnRowInView, boolean preserveViewportRow)
   {
      if (browseModel != null)
      {
         int viewportRow = getViewportRow();

         boolean editing = getActiveEditor() != null;
         if (editing)
         {
            undoValue();
            stopEdit();
         }

         int row = browseModel.getCurrentRow();
         if (row < 0)
         {
            // Protection code if the query has gone off-end for some reason.
            row = 0;
         }

         if (leaveOnRowInView && isRowInView(row))
         {
            // target row is in view, no need to scroll in this mode
            this.currentRow = row;
            currentRowAvail = true;
         }
         else
         {
            scrollToRow(row, true);
         }

         if (preserveViewportRow)
         {
            viewportRow = Math.min(viewportRow, getCacheSize() - 1);
            setCurrentRow(getCachedRowNumber(viewportRow), true);
         }

         if (editing)
            startEdit();

         repaint();
      }
   }

   /**
    * Refresh the browse after some of its rows were deleted.
    *
    * @param   rowIndex
    *          0-based index of the deleted row. <code>-1</code> if all of the selected rows were
    *          deleted.
    */
   public void refreshAfterDeletion(int rowIndex)
   {
      boolean singleRow = rowIndex != -1;

      if (!singleRow)
      {
         int numSelected = getSelectedRowsCount();

         if (numSelected == 1)
         {
            singleRow = true;
            rowIndex = getSelectedRowIndex(0);
         }
         else
         {
            int lastRow = getMaximalSelectedRowIndex();

            notifyRowDeleted(-1);

            boolean inView = isRowInView(lastRow) || isRowInView(lastRow + 1);

            if (!inView)
            {
               // scroll to the last row
               setCurrentRow(lastRow + 1 - numSelected, false);
            }
            // otherwise preserve the scroll row

            refreshScrollRow(true, true);

            if (inView)
            {
               refresh(true);
            }
         }
      }

      if (singleRow)
      {
         refresh(true);

         if (config.multiple)
         {
            notifyRowDeleted(rowIndex);
            setCurrentRow(rowIndex, false);
            refreshScrollRow(true, false);
         }
      }
   }
   
   /**
    * Reset the cache and redisplay.
    */
   public void queryClosed()
   {
      if (config.refreshable)
      {
         stopEdit();
         resetRowDisplayState();
         scrollToRow(0, true, false);
      }

      if (selectedRowIndices != null)
         selectedRowIndices.clear();

      rowCountAssessment = -1;
      repaint();
   }

   /**
    * Implements the SELECT-NEXT-ROW() widget method.<p>
    * Deselects all currently selected rows in a browse and selects the row after the deselected
    * row.<p>
    * This method also repositions the query to the new row and copies the record into the
    * database buffer. The browse automatically scrolls to the selected row if it is out of
    * view.<p>
    * This method is intended for use with a browse that supports the selection of a single row at
    * a time (MULTIPLE attribute is set to FALSE). If more than one row is selected when you
    * execute this method, all of the selected rows are deselected and the record after the last
    * selected row becomes the selected row.
    *
    * @return  <code>true</code> on success.
    */
   public boolean selectNextRow()
   {
      return advanceRowSelection(true);
   }

   /**
    * Implements the SELECT-PREV-ROW() widget method.<p>
    * Deselects all currently selected rows in a browse and selects the row before the deselected
    * row.<p>
    * This method also repositions the query to the new row and copies the record into the
    * database buffer.The browse automatically scrolls to the selected row if it is out of
    * view.<p>
    * This method is intended for use with a browse that supports the selection of a single row at
    * a time (MULTIPLE attribute is set to FALSE). If more than one row is selected when you
    * execute this method, all of the selected rows are deselected and the record before the last
    * selected row becomes the selected row.
    *
    * @return  <code>true</code> on success.
    */
   public boolean selectPreviousRow()
   {
      return advanceRowSelection(false);
   }

   /**
    * Notify the browse that some columns have been added to it.
    */
   public void notifyColumnsUpdated()
   {
      notifyColumnsUpdated(true);
   }

   /**
    * Notify the browse that some columns have been added to it.
    *
    * @param resetColumns
    *        If <code>true</code> columns cache in browse model is reset.
    */
   public void notifyColumnsUpdated(boolean resetColumns)
   {
      if (resetColumns)
      {
         browseModel.resetColumnsCache();
      }
      resetCacheStructures();
      resetRowDisplayState();

      dcData = null;
      setSize(originalWidth, originalHeight);
   }

   /**
    * Worker for {@link #selectNextRow()} and {@link #selectPreviousRow()}.
    *
    * @param  forward
    *         If <code>true</code> then it corresponds {@link #selectNextRow()}. If
    *         <code>false</code> - {@link #selectPreviousRow()}.
    *
    * @return <code>true</code> on success.
    */
   protected boolean advanceRowSelection(boolean forward)
   {
      boolean editing = getActiveEditor() != null;

      int selectedRow = getLastAvailableSelectedRow();
      if (selectedRow == -1)
      {
         // no *available* rows selected
         throw new ErrorConditionException(382, "");
      }

      BaseDataType[][] followingRow = null;
      if (!editing)
      {
         // no more *available* rows
         followingRow = forward ? browseModel.getRows(selectedRow + 1, 1, true):
               browseModel.getRows(selectedRow - 1, 1, false);
         if (followingRow.length == 0)
            return false;
      }

      boolean inView = isRowInView(selectedRow);

      // special case - read-only multi-selection, last available selected row is in view
      boolean specialCase = inView && selectedRowIndices != null && !editing;

      if (editing || inView)
      {
         int savedViewportRow = getViewportRow();
         boolean refresh      = false;
         if (!editing && isViewportEdge(selectedRow, forward))
         {
            refresh = true;
         }
         else
         {
            boolean res = advanceCurrRowOrSelection(specialCase, forward, editing, selectedRow);
            if (!editing && res)
            {
               // go down until we will find an available row
               while (specialCase ? !isSelectedRowAvailable(0) :
                                     getCurrentRow() != browseModel.getCurrentRow())
               {
                  selectedRow = getSelectedRowIndex(0, true);
                  if (isViewportEdge(selectedRow, forward))
                  {
                     // we have reached the end of view but didn't find an available row
                     refresh = true;
                     break;
                  }

                  if (!advanceCurrRowOrSelection(specialCase, forward, false, selectedRow))
                     break;
               }
            }
         }

         if (refresh)
         {
            if (!browseModel.isRowAvailable(getCurrentRow()))
            {
               // do not preserve viewport row if current row is not available
               savedViewportRow = 0;
            }

            if (forward)
            {
               // This is triggered in the following case:
               //    ...
               //  > current (available) row
               //    0 or more unavailable rows
               //    --- end of view ---
               scrollToRow(getCachedRowNumber(1));
               selectedRow = getCachedRowNumber(getCacheSize() - 1);
            }
            else
            {
               // This is triggered in the following case:
               //    ...
               //    --- begin of view ---
               //    0 or more unavailable rows
               //  > current (available) row
               int rowIndex = ((integer) followingRow[0][0]).intValue();
               scrollToRow(rowIndex);
               selectedRow = getCachedRowNumber(0);
            }

            if (specialCase)
            {
               selectedRowIndices.clear();
               selectedRowIndices.add(new SelectedRow(selectedRow));
            }
            else
            {
               setCurrentRow(selectedRow);
            }
         }

         if (specialCase)
         {
            // preserve viewport row
            setCurrentRow(getCachedRowNumber(Math.min(getCacheSize() - 1, savedViewportRow)));
         }
      }
      else
      {
         // multi-selection read-only browse, last available selected row is out of the view
         int rowIndex = ((integer) followingRow[0][0]).intValue();
         if (!isRowInView(rowIndex))
         {
            int viewportRow = getViewportRow();
            scrollToRow(rowIndex);
            // preserve viewport row
            setCurrentRow(getCachedRowNumber(Math.min(getCacheSize() - 1, viewportRow)));
         }
         selectedRowIndices.clear();
         selectedRowIndices.add(new SelectedRow(rowIndex));
      }

      repaint();
      return true;
   }

   /**
    * Determines if the specified row is the first or last row in the view.
    *
    * @param  rowIndex
    *         0-based row index in the browse model.
    * @param  forward
    *         <code>true</code> to check if it is the last row, <code>false</code> to check if it
    *         is the first row.
    *
    * @return <code>true</code> if the specified row is the first (if <code>forward</code> is
    *         <code>true</code>) or last (if <code>forward</code> is <code>false</code>) row in
    *         the view. If the page is not full then the last row on the page is NOT considered to
    *         be the last row in the view.
    */
   protected boolean isViewportEdge(int rowIndex, boolean forward)
   {
      return getViewportRow(rowIndex) == (forward ? ilPageRowsCount - 1 : 0);
   }

   /**
    * Works in two modes:
    * <ul>
    *    <li>Advance current row - make the row before or after the current row the new
    *    current row.</li>
    *    <li>Advance selection - deselected all rows and make the row before or after the
    *    specified row selected.</li>
    * </ul>
    *
    * @param  advanceSelection
    *         <code>true</code> for selection mode. <code>false</code> for current row mode.
    * @param  forward
    *         <code>true</code> to advance to the next row. <code>false</code> to advance to the
    *         previous row.
    * @param  handleEditing
    *         <code>true</code> to handle start/stop editing. For current row mode only.
    * @param  rowIndex
    *         0-based index of the row which should be used as a base for selection advancing (i.e.
    *         the row before or after it will be selected). For selection mode only.
    *
    * @return <code>true</code> if current/selected row has changed.
    */
   protected boolean advanceCurrRowOrSelection(boolean advanceSelection,
                                               boolean forward,
                                               boolean handleEditing,
                                               int rowIndex)
   {
      if (forward)
      {
         if (advanceSelection)
            return goDownWithSelection(rowIndex);
         else
            return goDown(handleEditing, false);
      }
      else
      {
         if (advanceSelection)
            return goUpWithSelection(rowIndex);
         else
            return goUp(handleEditing, false);
      }
   }

   /**
    * Get last (in the order of selection) "available" (not deleted on the moment of selection)
    * selected row.
    *
    * @return  0-based index of the selected row in the browse model or <code>-1</code> if there
    *          are no "available" selected rows.
    */
   protected int getLastAvailableSelectedRow()
   {
      if (selectedRowIndices == null)
      {
         SelectedRow row = getCurrentRowAsSelected();
         return row.isRowAvailable() ? row.getRowIndex() : -1;
      }
      else
      {
         int selectedRowsCount = selectedRowIndices.size();
         if (selectedRowsCount == 0)
            return -1;

         SelectedRow[] rows = selectedRowIndices.toArray(new SelectedRow[selectedRowsCount]);

         // find last available selected row
         for (int i = rows.length - 1; i >= 0; i--)
         {
            SelectedRow row = rows[i];
            if (row.isRowAvailable())
               return row.getRowIndex();
         }

         return -1;
      }
   }

   /**
    * Determines if the specified row is in the viewport.
    *
    * @param  rowIndex
    *         Index of the target row in the browse model.
    *
    * @return <code>true</code> if the specified row is in the viewport.
    */
   protected boolean isRowInView(int rowIndex)
   {
      int cacheSize = getCacheSize();
      for (int i = 0; i < cacheSize; i++)
      {
         if (getCachedRowNumber(i) == rowIndex)
         {
            return true;
         }
      }
      return false;
   }

   /**
    * Reset the state of the rows, as when no row has been displayed.
    */
   protected void resetRowDisplayState()
   {
      if (rowDisplayed != null)
      {
         Arrays.fill(rowDisplayed, false);
      }
      oldDisplayRow = -1;
   }

   /**
    * Get width of the browse area available for scrollable columns, i.e. browse page width minus
    * width of locked columns.
    *
    * @return see above.
    */
   protected int getAvailableWidth()
   {
      BrowseColumnConfig[] cols = browseModel.getColumns();

      int lockedWidth = 0;
      for (int i = 0; i < config.numLockedColumns; i++)
         lockedWidth += cols[i].widthChars + 1;

      return Math.max(0, (int) ilPageWidth - lockedWidth);
   }

   /**
    * Scroll one row up and adjust current row.
    *
    * @param  handleEditing
    *         If <code>true</code> then do {@link #startEdit}/{@link #stopEdit} calls.
    *
    * @return <code>true</code> if the current row has changed.
    */
   protected boolean goUp(boolean handleEditing)
   {
      return goUp(handleEditing, true);
   }

   /**
    * Scroll one row up and adjust current row.
    *
    * @param  handleEditing
    *         If <code>true</code> then do {@link #startEdit}/{@link #stopEdit} calls.
    * @param  fireValueChanged
    *         If <code>true</code> then VALUE-CHANGED event is fired if current row has changed.
    *
    * @return <code>true</code> if the current row has changed.
    */
   protected boolean goUp(boolean handleEditing, boolean fireValueChanged)
   {
      boolean editing = handleEditing && getActiveEditor() != null;

      if (editing && !leaveRow())
         return false;

      if (getCurrentRow() <= 0)
      {
         if (editing)
            startEdit();
         return false;
      }

      int viewportRow = getViewportRow();
      int newRow;
      if (viewportRow > 0)
      {
         newRow = getCachedRowNumber(viewportRow - 1);
      }
      else
      {
         if (!incrementalScroll(-1))
         {
            // no more rows
            if (editing)
               startEdit();
            return false;
         }

         newRow = getCachedRowNumber(0);
      }

      setCurrentRow(newRow);

      if (editing)
      {
         startEdit();

         if (fireValueChanged)
            fireValueChanged();
      }
      else if (!isSelectionMultiple() && fireValueChanged)
      {
         fireValueChanged();
      }

      return true;
   }

   /**
    * Scroll one row up and adjust current row.
    *
    * @param  handleEditing
    *         If <code>true</code> then do {@link #startEdit}/{@link #stopEdit} calls.
    *
    * @return <code>true</code> if the current row has changed.
    */
   protected boolean goDown(boolean handleEditing)
   {
      return goDown(handleEditing, true);
   }

   /**
    * Scroll one row down and adjust current row.
    *
    * @param handleEditing
    *        If <code>true</code> then do {@link #startEdit}/{@link #stopEdit} calls.
    * @param fireValueChanged
    *        If <code>true</code> then VALUE-CHANGED event is fired if current row has changed.
    *
    * @return <code>true</code> if the current row has changed.
    */
   protected boolean goDown(boolean handleEditing, boolean fireValueChanged)
   {
      boolean editing = handleEditing && getActiveEditor() != null;

      if (editing && !leaveRow())
         return false;

      int viewportRow = getViewportRow();
      int newRow;
      if (viewportRow < ilPageRowsCount - 1)
      {
         if (viewportRow == getCacheSize() - 1)
         {
            // we are on the last cached row
            if (editing)
               startEdit();
            return false;
         }

         newRow = getCachedRowNumber(viewportRow + 1);
      }
      else
      {
         if (!incrementalScroll(1))
         {
            // no more rows
            if (editing)
               startEdit();
            return false;
         }

         newRow = getCachedRowNumber(ilPageRowsCount - 1);
      }

      setCurrentRow(newRow);

      if (editing)
      {
         startEdit();

         if (fireValueChanged)
            fireValueChanged();
      }
      else if (!isSelectionMultiple() && fireValueChanged)
      {
         fireValueChanged();
      }

      return true;
   }

   /**
    * Deselect all rows and select the row before the given one.
    *
    * @param  rowIndex
    *         0-based index of the base row in the browse mode, the previous row before it will be
    *         selected.
    *
    * @return <code>true</code> if there are rows before the given one.
    */
   protected boolean goUpWithSelection(int rowIndex)
   {
      int viewportRow = getViewportRow(rowIndex);
      int newRow;
      boolean rowAvail;
      if (viewportRow > 0)
      {
         newRow = getCachedRowNumber(viewportRow - 1);
         rowAvail = browseModel.isRowAvailable(newRow);
      }
      else
      {
         if (!incrementalScroll(-1))
         {
            // no more rows
            return false;
         }

         newRow = getCachedRowNumber(0);
         rowAvail = true;
      }

      selectedRowIndices.clear();
      selectedRowIndices.add(new SelectedRow(newRow, rowAvail));
      return true;
   }

   /**
    * Deselect all rows and select the row after the given one.
    *
    * @param  rowIndex
    *         0-based index of the base row in the browse mode, the next row after it will be
    *         selected.
    *
    * @return <code>true</code> if there are rows after the given one.
    */
   protected boolean goDownWithSelection(int rowIndex)
   {
      int viewportRow = getViewportRow(rowIndex);
      int newRow;
      boolean rowAvail;
      if (viewportRow < ilPageRowsCount - 1)
      {
         if (viewportRow == getCacheSize() - 1)
         {
            // we are on the last cached row
            return false;
         }

         newRow = getCachedRowNumber(viewportRow + 1);
         rowAvail = browseModel.isRowAvailable(newRow);
      }
      else
      {
         if (!incrementalScroll(1))
         {
            // no more rows
            return false;
         }

         newRow = getCachedRowNumber(ilPageRowsCount - 1);
         rowAvail = true;
      }

      selectedRowIndices.clear();
      selectedRowIndices.add(new SelectedRow(newRow, rowAvail));
      return true;
   }

   /**
    * 4GL has two scrolling modes: "one-page" scrolling mode which is used if the total number
    * of rows is less or equal page size and the "normal" mode. In "one-page" mode scroll row
    * cannot be changed, the first row is the permanent scroll row.
    * 4GL can switch between these modes,
    * TODO switching can be implemented after gap closing mechanism for deleted rows is implemented
    *
    * @return <code>true</code> if "one-page" scrolling mode is active.
    */
   protected boolean isOnePageScrollingMode()
   {
      return (rowCountAssessment <= ilPageRowsCount) ||
             (getCachedRowNumber(0) == 0 && getCacheSize() < ilPageRowsCount);
   }

   /**
    * Scroll to the specified row (all records are refreshed).
    *
    * @param  scrollRow
    *         The new scroll row (i.e. the first row in the viewport).
    *
    * @return <code>true</code> if we got any rows from the server.
    */
   protected boolean scrollToRow(int scrollRow)
   {
      return scrollToRow(scrollRow, false, true);
   }

   /**
    * Scroll to the specified row (all records are refreshed).
    *
    * @param  scrollRow
    *         The new scroll row (i.e. the first row in the viewport).
    * @param  setAsCurrentRow
    *         Set the new scroll row as the current row.
    *
    * @return <code>true</code> if we got any rows from the server.
    */
   protected boolean scrollToRow(int scrollRow, boolean setAsCurrentRow)
   {
      return scrollToRow(scrollRow, setAsCurrentRow, false);
   }

   /**
    * Scroll to the specified row (all records are refreshed).
    *
    * @param  scrollRow
    *         The new scroll row (i.e. the first row in the viewport).
    * @param  setAsCurrentRow
    *         Set the new scroll row as the current row.
    * @param  leaveOnNoRows
    *         Return if there are no rows in the result set (otherwise cache will be cleared).
    *
    * @return <code>true</code> if we got any rows from the server.
    */
   protected boolean scrollToRow(int scrollRow, boolean setAsCurrentRow, boolean leaveOnNoRows)
   {
      return scrollToRow(scrollRow, setAsCurrentRow, leaveOnNoRows, false);
   }

   /**
    * Scroll to the specified row (all records are refreshed).
    *
    * @param  scrollRow
    *         The new scroll row (i.e. the first row in the viewport).
    * @param  setAsCurrentRow
    *         Set the new scroll row as the current row.
    * @param  leaveOnNoRows
    *         Return if there are no rows in the result set (otherwise cache will be cleared).
    * @param  ignoreOnePageScrollingMode
    *         Perform scroll as if the browse is in multiple-pages scrolling mode.
    *
    * @return <code>true</code> if we got any rows from the server.
    */
   protected boolean scrollToRow(int scrollRow,
                                 boolean setAsCurrentRow,
                                 boolean leaveOnNoRows,
                                 boolean ignoreOnePageScrollingMode)
   {
      if (browseModel == null)
      {
         dcData = null;
         setCurrentRow(0, true, true);
         resetCacheStructures();
         repaint();
         return false;
      }

      boolean firstScroll = rowCountAssessment < 0;
      boolean onePageScrolling = !ignoreOnePageScrollingMode && isOnePageScrollingMode();

      BaseDataType[][] serverData = browseModel.getRows(onePageScrolling ? 0 : scrollRow,
                                                        firstScroll ? ilPageRowsCount + 1 :
                                                                      ilPageRowsCount,
                                                        true);
      if (serverData.length > 0)
      {
         if (firstScroll)
         {
            dcData = new BaseDataType[Math.min(serverData.length, ilPageRowsCount)][];
            System.arraycopy(serverData, 0, dcData, 0, dcData.length);
            rowCountAssessment = serverData.length;
         }
         else
         {
            dcData = serverData;
         }

         if (setAsCurrentRow)
            setCurrentRow(scrollRow);
      }
      else
      {
         if (firstScroll)
         {
            rowCountAssessment = 0;
            dcData = null;
            setCurrentRow(0, true, true);
         }
         else
         {
            // we are beyond the end of the result set
            serverData = browseModel.getRows(scrollRow, 1, false);

            if (serverData.length == 0)
            {
               if (leaveOnNoRows)
                  return false;

               // the result set is empty
               dcData = null;
               setCurrentRow(0, true, true);
            }
            else
            {
               // we have scrolled to the last row
               dcData = serverData;
               if (setAsCurrentRow)
                  setCurrentRow(getCachedRowNumber(0));
            }
         }
      }

      resetCacheStructures();
      repaint();

      if (firstScroll &&
          editPossible &&
          getActiveEditor() == null &&
          getCacheSize() > 0)
      {
         startEdit();
      }

      return true;
   }

   /**
    * Reset ROW-DISPLAYED and COLOR arrays in order to reflect refreshed cache data.
    */
   protected void resetCacheStructures()
   {
      if (dcData != null)
      {
         rowDisplayed = new boolean[dcData.length];
         Arrays.fill(rowDisplayed, false);

         int cols = dcData[0] != null ? dcData[0].length : 0;
         colors = new Color[dcData.length][cols];

         rowDisplayTriggered = false;
      }
   }

   /**
    * Update height of the labels area.
    */
   protected void updateLabelsHeigth()
   {
      BrowseColumnConfig[] cols = browseModel.getColumns();
      for (BrowseColumnConfig col : cols)
      {
         ilLabelsHeight = Math.max(ilLabelsHeight, getLabelVar(col).length);
      }
   }

   /**
    * Calculate row index for internal cache.
    * 
    * @param   rowIndex
    *          Index in the model.
    *          
    * @return  Calculated index in the local cache.
    */
   private int dcGetRowIndex(int rowIndex)
   {
      rowIndex = Math.max(rowIndex, 0);
      
      int size = getCacheSize();
      for (int i = 0; i < size; i++)
      {
         if (getCachedRowNumber(i) == rowIndex)
            return i;
      }

      StringBuilder sb = new StringBuilder();
      for (int i = 0; i < size; i++)
      {
         sb.append(getCachedRowNumber(i));
         if (i != size - 1)
            sb.append(", ");
      }

      throw new IllegalArgumentException
            ("Invalid cache read request: " + rowIndex +
             ". Cache contains only [" + sb.toString() + "]");
   }

   /**
    * Invoke ROW-LEAVE trigger.
    *
    * @return <code>true</code> if there is no ROW-LEAVE trigger or it has returned normally.
    *         <code>false</code> if NO-APPLY has been returned by the trigger.
    */
   private boolean invokeRowLeaveTrigger()
   {
      int code = Keyboard.eventCode(Keyboard.ROW_LEAVE);
      ThinClient.TriggerResult res = ThinClient.getInstance().invokeTriggers(this, code, false);
      return !res.executed || !res.consume;
   }

   /**
    * Invoke ROW-LEAVE trigger and release the current row.
    *
    * @return <code>true</code> if ROW-LEAVE trigger returned normally (if it exists) and row
    *         release was successful.
    */
   private boolean leaveRow()
   {
      if (!invokeRowLeaveTrigger())
         return false;

      if (!releaseCurrentRow())
      {
         startEdit();
         return false;
      }

      return true;
   }

   /**
    * Container which holds row index and row availability (i.e. if it could be fetched) at the
    * moment of selection.
    */
   private static class SelectedRow
   {
      /** Row index. */
      private final int rowIndex;

      /** Row availability. */
      private final boolean rowAvailable;

      /**
       * Creates a new selected row which is available.
       *
       * @param rowIndex
       *        Row index.
       */
      public SelectedRow(int rowIndex)
      {
         this.rowIndex = rowIndex;
         this.rowAvailable = true;
      }

      /**
       * Creates a new selected row.
       *
       * @param rowIndex
       *        Row index.
       * @param rowAvailable
       *        Row availability.
       */
      public SelectedRow(int rowIndex, boolean rowAvailable)
      {
         this.rowIndex = rowIndex;
         this.rowAvailable = rowAvailable;
      }

      /**
       * Get row index.
       *
       * @return row index.
       */
      public int getRowIndex()
      {
         return rowIndex;
      }

      /**
       * Get row availability.
       *
       * @return <code>true</code> if the row is available (i.e. if it could be fetched) at the
       *         moment of selection.
       */
      public boolean isRowAvailable()
      {
         return rowAvailable;
      }

      /**
       * Return a hash code for the instance.
       *
       * @return  Hash code for the instance.
       */
      @Override
      public int hashCode()
      {
         return rowIndex;
      }

      /**
       * Compares two objects for equality. Two selected rows are considered equal if they have
       * the same row index (regardless of their availability).
       *
       * @param   obj
       *          Object to compare with.
       *
       * @return  <code>true</code> if instances are equal and <code>false</code> otherwise.
       */
      @Override
      public boolean equals(Object obj)
      {
         if (this == obj)
            return true;

         if (obj == null || getClass() != obj.getClass())
            return false;

         SelectedRow other = (SelectedRow) obj;
         return other.rowIndex == this.rowIndex;
      }
   }
}
